from dipy.align import resample
from dipy.segment.clustering import QuickBundles
from dipy.segment.metric import AveragePointwiseEuclideanMetric
from dipy.segment.featurespeed import ResampleFeature
from dipy.io.streamline import (
    load_tractogram, save_tractogram, StatefulTractogram, Space)
from dipy.data.fetcher import _make_fetcher
import dipy.data as dpd
from AFQ.utils.path import drop_extension, apply_cmd_to_afq_derivs

import os
import os.path as op
import json
from glob import glob
import shutil

import numpy as np
import pandas as pd
import logging
import time
from tqdm import tqdm

import warnings
import nibabel as nib
import boto3
from botocore import UNSIGNED
from botocore.client import Config


# capture templateflow resource warning and log
default_warning_format = warnings.formatwarning
try:
    warnings.formatwarning = lambda msg, *args, **kwargs: f'{msg}'
    logging.captureWarnings(True)
    pywarnings_logger = logging.getLogger('py.warnings')
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
    pywarnings_logger.addHandler(console_handler)

    warnings.filterwarnings(
        "default", category=ResourceWarning,
        module="templateflow")

    from templateflow import api as tflow
finally:
    logging.captureWarnings(False)
    warnings.formatwarning = default_warning_format


__all__ = ["fetch_callosum_templates", "read_callosum_templates",
           "fetch_or_templates", "read_or_templates",
           "fetch_templates", "read_templates",
           "fetch_stanford_hardi_tractography",
           "read_stanford_hardi_tractography",
           "organize_stanford_data",
           "fetch_stanford_hardi_lv1"]


# Set a user-writeable file-system location to put files:
if 'AFQ_HOME' in os.environ:
    afq_home = os.environ['AFQ_HOME']
else:
    afq_home = op.join(op.expanduser('~'), 'AFQ_data')

baseurl = "https://ndownloader.figshare.com/files/"


def _make_reusable_fetcher(name, folder, baseurl, remote_fnames, local_fnames,
                           doc="", md5_list=None, **make_fetcher_kwargs):
    def fetcher():
        all_files_downloaded = True
        for fname in local_fnames:
            if not op.exists(op.join(folder, fname)):
                all_files_downloaded = False
        if all_files_downloaded:
            files = {}
            for i, (f, n), in enumerate(zip(remote_fnames, local_fnames)):
                files[n] = (baseurl + f, md5_list[i] if
                            md5_list is not None else None)
            return files, folder
        else:
            return _make_fetcher(
                name, folder, baseurl, remote_fnames, local_fnames,
                doc=doc, **make_fetcher_kwargs)()
    fetcher.__name__ = name
    fetcher.__doc__ = doc
    return fetcher


def _fetcher_to_template(fetcher, as_img=False, resample_to=False):
    if isinstance(resample_to, str):
        resample_to = nib.load(resample_to)
    files, folder = fetcher()
    template_dict = {}
    for f in files:
        img = op.join(folder, f)
        if as_img:
            img = nib.load(img)
        if resample_to:
            img = nib.Nifti1Image(resample(img.get_fdata(),
                                           resample_to,
                                           img.affine,
                                           resample_to.affine).get_fdata(),
                                  resample_to.affine)
        template_dict[drop_extension(f)] = img
    return template_dict


callosum_fnames = ["Callosum_midsag.nii.gz",
                   "L_AntFrontal.nii.gz",
                   "L_Motor.nii.gz",
                   "L_Occipital.nii.gz",
                   "L_Orbital.nii.gz",
                   "L_PostParietal.nii.gz",
                   "L_SupFrontal.nii.gz",
                   "L_SupParietal.nii.gz",
                   "L_Temporal.nii.gz",
                   "R_AntFrontal.nii.gz",
                   "R_Motor.nii.gz",
                   "R_Occipital.nii.gz",
                   "R_Orbital.nii.gz",
                   "R_PostParietal.nii.gz",
                   "R_SupFrontal.nii.gz",
                   "R_SupParietal.nii.gz",
                   "R_Temporal.nii.gz"]

callosum_remote_fnames = ["5273794", "5273797", "5273800", "5273803",
                          "5273806", "5273809", "5273812", "5273815",
                          "5273821", "5273818", "5273824", "5273827",
                          "5273830", "5273833", "5273836", "5273839",
                          "5273842"]

callosum_md5_hashes = ["709fa90baadeacd64f1d62b5049a4125",
                       "987c6169de807c4e93dc2cbd7a25d506",
                       "0da114123d0b0097b96fe450a459550b",
                       "6d845bd10504f67f1dc17f9000076d7e",
                       "e16c7873ef4b08d26b77ef746dab8237",
                       "47193fd4df1ea17367817466de798b90",
                       "7e78bf9671e6945f4b2f5e7c30595a3c",
                       "8adbb947377ff7b484c88d8c0ffc2125",
                       "0fd981a4d0847e0642ff96e84fe44e47",
                       "87c4855efa406d8fb004cffb8259180e",
                       "c7969bcf5f2343fd9ce9c49b336cf14c",
                       "bb4372b88991932150205ffb22aa6cb7",
                       "d198d4e7db18ddc7236cf143ecb8342e",
                       "d0f6edef64b0c710c92e634496085dda",
                       "85eaee44665f244db5adae2e259833f6",
                       "25f24eb22879a05d12bda007c81ea55a",
                       "2664e0b8c2d9c59f13649a89bfcce399"]

fetch_callosum_templates = _make_reusable_fetcher(
    "fetch_callosum_templates",
    op.join(afq_home,
            'callosum_templates'),
    baseurl, callosum_remote_fnames,
    callosum_fnames,
    md5_list=callosum_md5_hashes,
    doc="Download AFQ callosum templates")


def read_callosum_templates(as_img=True, resample_to=False):
    """Load AFQ callosum templates from file

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict with: keys: names of template ROIs and values: nibabel Nifti1Image
    objects from each of the ROI nifti files.
    """
    logger = logging.getLogger('AFQ')

    logger.debug('loading callosum templates')
    tic = time.perf_counter()

    template_dict = _fetcher_to_template(
        fetch_callosum_templates,
        as_img=as_img,
        resample_to=resample_to)

    toc = time.perf_counter()
    logger.debug(f'callosum templates loaded in {toc - tic:0.4f} seconds')

    return template_dict


##########################################################################
# Pediatric templates:
# --------------------
#
# Templates include the UNC 0-1-2 Infant Atlases and waypoint ROIs matched
# to the atlas.
#
# Information on:
#
# - atlas: https://www.nitrc.org/projects/pediatricatlas
#
# - pediatric templates: https://github.com/yeatmanlab/AFQ/tree/babyAFQ
#
# Templates downloaded from:
#
# - https://figshare.com/articles/dataset/ROIs_probabilistic_maps_and_transform_data_for_pediatric_automated_fiber_quantification/13027487  # noqa
#

pediatric_fnames = [
    "ATR_roi1_L.nii.gz", "ATR_roi1_R.nii.gz",
    "ATR_roi2_L.nii.gz", "ATR_roi2_R.nii.gz",
    "ATR_roi3_L.nii.gz", "ATR_roi3_R.nii.gz",
    "CGC_roi1_L.nii.gz", "CGC_roi1_R.nii.gz",
    "CGC_roi2_L.nii.gz", "CGC_roi2_R.nii.gz",
    "CGC_roi3_L.nii.gz", "CGC_roi3_R.nii.gz",
    "CST_roi1_L.nii.gz", "CST_roi1_R.nii.gz",
    "CST_roi2_L.nii.gz", "CST_roi2_R.nii.gz",
    "FA_L.nii.gz", "FA_R.nii.gz",
    "FP_L.nii.gz", "FP_R.nii.gz",
    "HCC_roi1_L.nii.gz", "HCC_roi1_R.nii.gz",
    "HCC_roi2_L.nii.gz", "HCC_roi2_R.nii.gz",
    "IFO_roi1_L.nii.gz", "IFO_roi1_R.nii.gz",
    "IFO_roi2_L.nii.gz", "IFO_roi2_R.nii.gz",
    "IFO_roi3_L.nii.gz", "IFO_roi3_R.nii.gz",
    "ILF_roi1_L.nii.gz", "ILF_roi1_R.nii.gz",
    "ILF_roi2_L.nii.gz", "ILF_roi2_R.nii.gz",
    "LH_Parietal.nii.gz", "RH_Parietal.nii.gz",
    "MdLF_roi1_L.nii.gz", "MdLF_roi1_R.nii.gz",
    "SLF_roi1_L.nii.gz", "SLF_roi1_R.nii.gz",
    "SLF_roi2_L.nii.gz", "SLF_roi2_R.nii.gz",
    "SLFt_roi2_L.nii.gz", "SLFt_roi2_R.nii.gz",
    "SLFt_roi3_L.nii.gz", "SLFt_roi3_R.nii.gz",
    "UNC_roi1_L.nii.gz", "UNC_roi1_R.nii.gz",
    "UNC_roi2_L.nii.gz", "UNC_roi2_R.nii.gz",
    "UNC_roi3_L.nii.gz", "UNC_roi3_R.nii.gz",
    "VOF_box_L.nii.gz", "VOF_box_R.nii.gz",
    "UNCNeo-withCerebellum-for-babyAFQ.nii.gz",
    "UNCNeo_JHU_tracts_prob-for-babyAFQ.nii.gz",
    "mid-saggital.nii.gz",
    "OR_rightV1.nii.gz",
    "OR_rightThal.nii.gz",
    "OR_right_roi3.nii.gz",
    "OR_leftV1.nii.gz",
    "OR_leftThal.nii.gz",
    "OR_left_roi3.nii.gz",
    "pARC_R_start.nii.gz",
    "pARC_L_start.nii.gz",
    "VOF_box_small_L.nii.gz",
    "VOF_box_small_R.nii.gz",
    "VOF_L_start.nii.gz",
    "VOF_R_start.nii.gz",
]

pediatric_md5_hashes = [
    "2efe0deb19ac9e175404bf0cb29d9dbd", "c2e07cd50699527bd7b0cbbe88703c56",
    "76b36d8d6759df58131644281ed16bd2", "645102225bad33da30bafd41d24b3ab0",
    "45ec94d42fdc9448afa6760c656920e9", "54e3cb1b8c242be279f1410d8bb3c383",
    "1ee9f7e8b21ef8ceee81d5a7a178ef33", "4f11097f7ae317aa8d612511be79e2f1",
    "1c4c0823c23b676d6d35004d93b9c695", "d4830d558cc8f707ebec912b32d197a5",
    "c405e0dbd9a4091c77b3d1ad200229b4", "ec0aeccc6661d2ee5ed79259383cdcee",
    "2802cd227b550f6e85df0fec1d515c29", "385addb999dc6d76957d2a35c4ee74bb",
    "0dd14c02b272263adbe2246880979c9d", "99dac5a00c10d81d222f020162fd6194",
    "e49ba370edca96734d9376f551d413db", "f59e9e69e06325198f70047cd63c3bdc",
    "ae3bd2931f95adae0280a8f75cd3ca9b", "c409a0036b8c2dd4d03d11fbc6bfbdcd",
    "c2597a474ea5ec9e3126c35fd238f6b2", "67af59c934147c9f9ff6e0b76c4cc6eb",
    "72d0bbc0b6162e9291fdc450c103a1f0", "51f5041be63ad0ac10d1ac09e3bf1a8e",
    "6200f5cdc1402dce46dedd505468b147", "83cb5bf6b9b1eda63c8643871e84a6d4",
    "2a5d8d309b1256d6e48958e112201c2c", "ba24d0915fdff215a403480d0a7745c9",
    "1001e833d1344274d543ffd02a66af80", "03e20c5eebcd4d439c4ffb36d26a10d9",
    "6200f5cdc1402dce46dedd505468b147", "83cb5bf6b9b1eda63c8643871e84a6d4",
    "a6ae325cce2dc4bb21b52ee4c6ca844f", "a96a31df8f96ccf1c5f871a18e4a2d72",
    "65b7378ca85689a5f23b1b84a6ed78f0", "ce0d0ea696ef51c671aa118e11192e2d",
    "ce4918086ca1e829698576bcf95db62b", "96168d2eff74ec2acf9065f499732526",
    "6b20ba319d44472ec21d6ece937605bb", "26b1cf6ec8bd365dde42e3efe9beeac2",
    "0b3ccf06564d973bfcfff9a87e74f8b5", "84f3426033a2b225b0920b2622864375",
    "5351b3cb7efa9aa8e83e266342809ebe", "4e8a34aaba4e0f22a6149f38620dc39d",
    "682c08f66e8c2cf9e4e60f5ce308c76c", "9077affd4f3a8a1d6b44678cde4b3bf4",
    "5adf36f00669cc547d5eb978acf46266", "66a8002688ffdf3943722361da90ec6a",
    "efb5ae138df92019541861f9aa6a4d57", "757ec61078b2e9f9a073871b3216ff7a",
    "ff1e238c52a21f8cc5d44ac614d9627f", "cf16dd2767c6ab2d3fceb2890f6c3e41",
    "6016621e244b60b9c69fd44b055e4a03", "fd495a2c94b6b13bfb4cd63e293d3fc0",
    "bf81a23d80f55e5f1eb0c16717193105",
    "6f8bf8f70216788d14d9a49a3c664b16",
    "19df0297d6a2ac21da5e432645d63174",
    "b4ffb957a2adbb8b76966e4ea28dbdf1",
    "3de1bc5aae4c76164f34515e2e84939c",
    "705ba1cbfc42ce64a54dda2b732f81f2",
    "bc6d4f880d3eb281358af1c764779704",
    "a3b6a7be067aa12af273482baee1498d",
    "f035813314960eb91f3a4dae508a68e5",
    "728461f81fcfa8f02ff3af969ab6499c",
    "5c1c87ee71c512b38a0711b93bb7e9fe",
    "4652bcf7a1f2b7cdeec52956eb884795",
    "e1c4bf76d2c98dcf6ffe00399a6e8b07",
    "0ecabd68fa9c56614e2a694359c0a545",
    "4c93bea7a72ac2b7475201acef5a1fc4",
]

pediatric_remote_fnames = [
    "24880625", "24880628", "24880631", "24880634", "24880637", "24880640",
    "24880643", "24880646", "24880649", "24880652", "24880655", "24880661",
    "24880664", "24880667", "46407571", "46407568", "24880676", "24880679",
    "24880685", "24880688", "24880691", "24880694", "24880697", "24880700",
    "24880703", "24880706", "24880712", "24880715", "24880718", "24880721",
    "24880724", "24880727", "24880730", "24880733", "24880736", "24880748",
    "24880739", "24880742", "24880754", "24880757", "24880760", "24880763",
    "24880769", "24880772", "24880775", "24880778", "24880781", "24880787",
    "24880790", "24880793", "24880796", "24880802", "24880805", "24880808",
    "24880616", "24880613", "24986396",
    "42120480",
    "42120483",
    "42120486",
    "42120489",
    "42120492",
    "42120495",
    "42121791",
    "42121794",
    "42121836",
    "42121839",
    "42121803",
    "42121806",
]

fetch_pediatric_templates = _make_reusable_fetcher(
    'fetch_pediatric_templates',
    op.join(afq_home, 'pediatric_templates'),
    'https://ndownloader.figshare.com/files/',
    pediatric_remote_fnames,
    pediatric_fnames,
    md5_list=pediatric_md5_hashes,
    doc='Download pediatric templates'
)


def read_pediatric_templates(as_img=True, resample_to=False):
    """
    Load pediatric pyAFQ templates.

    Used to create pediatric `bundle_dict`.

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict :
        keys = names of template ROIs, and
        values = `Nifti1Image` from each of the ROI nifti files.
    """

    print('Loading pediatric templates...', flush=True)
    pediatric_templates = _fetcher_to_template(
        fetch_pediatric_templates,
        as_img=as_img,
        resample_to=resample_to)

    # For the arcuate (AF/ARC), reuse the SLF ROIs
    pediatric_templates['ARC_roi1_L'] = pediatric_templates['SLF_roi1_L']
    pediatric_templates['ARC_roi1_R'] = pediatric_templates['SLF_roi1_R']
    pediatric_templates['ARC_roi2_L'] = pediatric_templates['SLFt_roi3_L']
    pediatric_templates['ARC_roi2_R'] = pediatric_templates['SLFt_roi3_R']
    pediatric_templates['ARC_roi3_L'] = pediatric_templates['SLFt_roi2_L']
    pediatric_templates['ARC_roi3_R'] = pediatric_templates['SLFt_roi2_R']

    # For the middle longitudinal fasciculus (MdLF) reuse ILF ROI
    pediatric_templates['MdLF_roi2_L'] = pediatric_templates['ILF_roi2_L']
    pediatric_templates['MdLF_roi2_R'] = pediatric_templates['ILF_roi2_R']

    return pediatric_templates


def read_resample_roi(roi, resample_to=None, threshold=False):
    """
    Reads an roi from file-name/img and resamples it to conform with
    another file-name/img.

    Parameters
    ----------
    roi : str or nibabel image class instance.
        Should contain a binary volume with 1s in the region of interest and
        0s elsewhere.

    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template.

    threshold: bool or float
        If set to False (default), resampled result is returned. Otherwise,
        the resampled result is thresholded at this value and binarized.
        This is not applied if the input ROI is already in the space of the
        output.

    Returns
    -------
    nibabel image class instance that contains the binary ROI resampled into
    the requested space.
    """
    if isinstance(roi, str):
        roi = nib.load(roi)

    if resample_to is None:
        resample_to = read_mni_template()

    if isinstance(resample_to, str):
        resample_to = nib.load(resample_to)

    if resample_to is False or np.allclose(resample_to.affine, roi.affine):
        return roi

    as_array = resample(
        roi.get_fdata(),
        resample_to,
        roi.affine,
        resample_to.affine).get_fdata()
    if threshold:
        as_array = (as_array > threshold).astype(int)

    img = nib.Nifti1Image(
        as_array,
        resample_to.affine)

    return img


template_fnames = ["ATR_roi1_L.nii.gz",
                   "ATR_roi1_R.nii.gz",
                   "ATR_roi2_L.nii.gz",
                   "ATR_roi2_R.nii.gz",
                   "ATR_L_prob_map.nii.gz",
                   "ATR_R_prob_map.nii.gz",
                   "CGC_roi1_L.nii.gz",
                   "CGC_roi1_R.nii.gz",
                   "CGC_roi2_L.nii.gz",
                   "CGC_roi2_R.nii.gz",
                   "CGC_L_prob_map.nii.gz",
                   "CGC_R_prob_map.nii.gz",
                   "CST_roi1_L.nii.gz",
                   "CST_roi1_R.nii.gz",
                   "CST_roi2_L.nii.gz",
                   "CST_roi2_R.nii.gz",
                   "CST_L_prob_map.nii.gz",
                   "CST_R_prob_map.nii.gz",
                   "FA_L.nii.gz",
                   "FA_R.nii.gz",
                   "FA_prob_map.nii.gz",
                   "FP_L.nii.gz",
                   "FP_R.nii.gz",
                   "FP_prob_map.nii.gz",
                   "HCC_roi1_L.nii.gz",
                   "HCC_roi1_R.nii.gz",
                   "HCC_roi2_L.nii.gz",
                   "HCC_roi2_R.nii.gz",
                   "HCC_L_prob_map.nii.gz",
                   "HCC_R_prob_map.nii.gz",
                   "IFO_roi1_L.nii.gz",
                   "IFO_roi1_R.nii.gz",
                   "IFO_roi2_L.nii.gz",
                   "IFO_roi2_R.nii.gz",
                   "IFO_L_prob_map.nii.gz",
                   "IFO_R_prob_map.nii.gz",
                   "ILF_roi1_L.nii.gz",
                   "ILF_roi1_R.nii.gz",
                   "ILF_roi2_L.nii.gz",
                   "ILF_roi2_R.nii.gz",
                   "ILF_L_prob_map.nii.gz",
                   "ILF_R_prob_map.nii.gz",
                   "SLF_roi1_L.nii.gz",
                   "SLF_roi1_R.nii.gz",
                   "SLF_roi2_L.nii.gz",
                   "SLF_roi2_R.nii.gz",
                   "SLFt_roi2_L.nii.gz",
                   "SLFt_roi2_R.nii.gz",
                   "SLF_L_prob_map.nii.gz",
                   "SLF_R_prob_map.nii.gz",
                   "UNC_roi1_L.nii.gz",
                   "UNC_roi1_R.nii.gz",
                   "UNC_roi2_L.nii.gz",
                   "UNC_roi2_R.nii.gz",
                   "UNC_L_prob_map.nii.gz",
                   "UNC_R_prob_map.nii.gz",
                   "ARC_L_prob_map.nii.gz",
                   "ARC_R_prob_map.nii.gz",
                   "VOF_R_end.nii.gz",
                   "VOF_R_start.nii.gz",
                   "VOF_L_end.nii.gz",
                   "VOF_L_start.nii.gz",
                   "pARC_R_start.nii.gz",
                   "pARC_L_start.nii.gz",
                   "ARC_R_end.nii.gz",
                   "ARC_R_start.nii.gz",
                   "ARC_L_end.nii.gz",
                   "ARC_L_start.nii.gz",
                   "UNC_R_end.nii.gz",
                   "UNC_R_start.nii.gz",
                   "UNC_L_end.nii.gz",
                   "UNC_L_start.nii.gz",
                   "SLF_R_end.nii.gz",
                   "SLF_R_start.nii.gz",
                   "SLF_L_end.nii.gz",
                   "SLF_L_start.nii.gz",
                   "ILF_R_end.nii.gz",
                   "ILF_R_start.nii.gz",
                   "ILF_L_end.nii.gz",
                   "ILF_L_start.nii.gz",
                   "IFO_R_end.nii.gz",
                   "IFO_R_start.nii.gz",
                   "IFO_L_end.nii.gz",
                   "IFO_L_start.nii.gz",
                   "FA_end.nii.gz",
                   "FA_start.nii.gz",
                   "FP_end.nii.gz",
                   "FP_start.nii.gz",
                   "CGC_R_start.nii.gz",
                   "CGC_L_start.nii.gz",
                   "CST_R_end.nii.gz",
                   "CST_R_start.nii.gz",
                   "CST_L_end.nii.gz",
                   "CST_L_start.nii.gz",
                   "ATR_R_end.nii.gz",
                   "ATR_R_start.nii.gz",
                   "ATR_L_end.nii.gz",
                   "ATR_L_start.nii.gz"]


template_remote_fnames = ["5273680", "5273683", "5273686", "5273689",
                          "11458274", "11458277",
                          "5273695", "5273692", "5273698", "5273701",
                          "11458268", "11458271",
                          "5273704", "5273707", "46407574", "46407577",
                          "11458262", "11458265",
                          "5273716", "5273719",
                          "11458220",
                          "5273722", "5273725",
                          "11458226",
                          "5273728", "5273731", "5273734", "5273746",
                          "11458259", "11458256",
                          "5273737", "5273740", "5273743", "5273749",
                          "11458250", "11458253",
                          "5273752", "5273755", "5273758", "5273761",
                          "11458244", "11458247",
                          "5273764", "5273767", "5273770", "5273773",
                          "5273776", "5273791",
                          "11458238", "11458241",
                          "5273779", "5273782", "5273785", "5273788",
                          "11458223", "11458229",
                          "11458232", "11458235",
                          "40943957",
                          "40943960",
                          "40943966",
                          "40943969",
                          "40943972",
                          "40943975",
                          "40943978",
                          "40943981",
                          "40943984",
                          "40943987",
                          "40943990",
                          "40943993",
                          "40943996",
                          "40943999",
                          "40944002",
                          "40944005",
                          "40944008",
                          "40944011",
                          "40944014",
                          "40944017",
                          "40944020",
                          "40944023",
                          "40944026",
                          "40944029",
                          "40944032",
                          "40944035",
                          "40944038",
                          "40944041",
                          "40944044",
                          "40944047",
                          "40944050",
                          "40944053",
                          "40944056",
                          "40944059",
                          "40944062",
                          "40944065",
                          "40944068",
                          "40944074",
                          "40944077",
                          "40944080"]


template_md5_hashes = ["6b7aaed1a2982fd0ea436a223133908b",
                       "fd60d46d4e3cbd906c86e4c9e4fd6e2a",
                       "3aba60b169a35c38640de4ec29d362c8",
                       "12716a5688a1809fbaed1d58d2e68b59",
                       "c5637f471df861d9bbb45604db34770b",
                       "850cc4c04d7241747063fe3cd440b2ce",
                       "8e8973bc7838c8744914d402f52d91ca",
                       "c5fa4e6e685e695c006823b6784d2407",
                       "e1fab77f21d5303ed52285f015e24f0b",
                       "5f89defec3753fd75cd688c7bfb20a36",
                       "a4f3cd65b06fb25f63d5dab7592f00f2",
                       "7e73ab02db30a3ad6bd9e82148c2486e",
                       "f9db3154955a20b67c2dda758800d14c",
                       "73941510c798c1ed1b03e2bd481cd5c7",
                       "b20e0caa54cf35002cd06cf6033b964f",
                       "e751306df304af32c3ce7617913bbd30",
                       "fd012bc89f6bed7bd54530195496bac4",
                       "3406906a86e633cc102127cf210a1063",
                       "9040a7953dcbbf131d135c866182d8ef",
                       "a72e17194824fcd838a594a2eb50c72e",
                       "627d7bb2e6d55f8243da815a36d9ff1a",
                       "55adbe9b8279185eedbe342149e1ff90",
                       "5a7412a3cf0fb185eec53d1989df2f7c",
                       "1aa36e83ae7b5555bb19d776ede9c18d",
                       "ba453196ff179b0e31172806e313b52c",
                       "d85c6574526b296935f34bf4f65cd493",
                       "9b81646317f59c7db087f27e2f85679e",
                       "9806e82c250e4604534b96917f87b7e8",
                       "213d3fb1ccd756d878f9b50b765b1c8f",
                       "f1e7e6bc51aa0aa279c54fb3805fb5e3",
                       "0e68a9feaaddcc9b4d667c2f15903368",
                       "d45020a87ee4bb496edd350631d91f6a",
                       "75c2c911826ec4b23159f9bd80e3c039",
                       "55d616ea9e0c646adc1aafa0f5fbe625",
                       "dee83fa6b03cfa5e0f5c965953aa6778",
                       "a13eef7059c98568adfefbab660e434e",
                       "045b7d5c6341997f3f0120c3a4212ad8",
                       "d174b1359ba982b03840436c93b7bbb4",
                       "fff9753f394fc4c73fb2ae40b3b4dde0",
                       "cd278b4dd6ff77481ea9ac16485a5ae2",
                       "7bdf5111265107091c7a2fca9215de30",
                       "7d4a43714504e6e930f922c9bc2a13d5",
                       "af2bcedf47e193686af329b9a8e259da",
                       "9a1122943579d11ba169d3ad87a75625",
                       "627903f7a06627bfd4153dc9245fa390",
                       "1714cd7f989c3435bdd5a2076e6272a0",
                       "1fa2114049707a4e05b53f9d95730375",
                       "b6663067d5ea53c70cb8803948f8adf7",
                       "d3e068997ebc60407bd6e9576e47dede",
                       "27ecfbd1d2f98213e52d73b7d70fe0e7",
                       "fa141bb2d951bec486916acda3652d95",
                       "d391d073e86e28588be9a6d01b2e7a82",
                       "a3e085562e6b8111c7ebc358f9450c8b",
                       "d65c67910807504735e034f7ea92d590",
                       "93cb24a9128db1a6c34a09eaf79fe7f0",
                       "71a7455cb4062dc39b1677c118c7b5a5",
                       "19590c712f1776da1fdba64d4eb7f1f6",
                       "04d5af0feb2c1b5b52a87ccbbf148e4b",
                       "53c277be990d00f7de04f2ea35e74d73",
                       "d37d815fd1bdaaf3a9d2dcfc3ccb1345",
                       "95ed3189d8ac152945e6be1eb24381a3",
                       "a9007e6f2d6ae13ef182f65057c06573",
                       "c6eb9ee33b7caf691749e266f89e8ec4",
                       "a06b2e2e52c09a601f683dc39859a7f1",
                       "bee876a34fdb03e69a418b791f90975a",
                       "680749c9e4565bc02492019d57d8e7d7",
                       "ffc157e9f73a43eff23821f2cfca614a",
                       "a92beacd59ff2c90408edc7407a571c4",
                       "1c0b570bb2d622718b01ee2c429a5d15",
                       "4d3db603f7cc86b5696b1514c5b43e0e",
                       "96c644616e8724b4d071c64981363e6c",
                       "9be681028226aa9b70fac46f8e376282",
                       "e19274bd1ee58906c03f5f9a7c4f01c5",
                       "1d33ac68aeb8eb444c07b5a853414727",
                       "ca3c698154d77217c9480123a49c73dd",
                       "b17d876c0ca51f0e3ec6177382aa4ecf",
                       "6ff7e2dd357affc7c4e4d79f5f5866e9",
                       "96c644616e8724b4d071c64981363e6c",
                       "de045095d914d4929409e062fc1a09b6",
                       "e19274bd1ee58906c03f5f9a7c4f01c5",
                       "69009da4aa52607962ec99b9e248a39f",
                       "2b106da3ed88e4cc9a5b4c391cf3445d",
                       "de045095d914d4929409e062fc1a09b6",
                       "0d28a8e367646f9f0036ba7a100e49e5",
                       "69009da4aa52607962ec99b9e248a39f",
                       "1c0b570bb2d622718b01ee2c429a5d15",
                       "ffc157e9f73a43eff23821f2cfca614a",
                       "69009da4aa52607962ec99b9e248a39f",
                       "de045095d914d4929409e062fc1a09b6",
                       "9d897a3ab960931c5119a9bbfc6d7838",
                       "0046be49e39cf8c639418c805de8b9e2",
                       "67d8f24bd749d8a9094ccb9d8e63fa25",
                       "5736d28551296253f879c4b0d450b70e",
                       "67d8f24bd749d8a9094ccb9d8e63fa25",
                       "5736d28551296253f879c4b0d450b70e",
                       "5f2c0c0f1b7b32ba0a8474a460a2b980",
                       "ffc157e9f73a43eff23821f2cfca614a",
                       "a8d308a93b26242c04b878c733cb252f",
                       "1c0b570bb2d622718b01ee2c429a5d15"]

fetch_templates = _make_reusable_fetcher(
    "fetch_templates",
    op.join(afq_home, 'templates'),
    baseurl, template_remote_fnames,
    template_fnames, md5_list=template_md5_hashes,
    doc="Download AFQ templates")


def read_templates(as_img=True, resample_to=False):
    """Load AFQ templates from file

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict with: keys: names of template ROIs and values: nibabel Nifti1Image
    objects from each of the ROI nifti files.
    """
    logger = logging.getLogger('AFQ')
    logger.debug('loading AFQ templates')
    tic = time.perf_counter()

    template_dict = _fetcher_to_template(
        fetch_templates,
        as_img=as_img,
        resample_to=resample_to)

    toc = time.perf_counter()
    logger.debug(f'AFQ templates loaded in {toc - tic:0.4f} seconds')

    return template_dict


cp_fnames = [
    "ICP_L_inferior_prob.nii.gz",
    "ICP_L_superior_prob.nii.gz",
    "ICP_R_inferior_prob.nii.gz",
    "ICP_R_superior_prob.nii.gz",
    "MCP_L_inferior_prob.nii.gz",
    "MCP_L_superior_prob.nii.gz",
    "MCP_R_inferior_prob.nii.gz",
    "MCP_R_superior_prob.nii.gz",
    "SCP_L_inferior_prob.nii.gz",
    "SCP_L_inter_prob.nii.gz",
    "SCP_L_superior_prob.nii.gz",
    "SCP_R_inferior_prob.nii.gz",
    "SCP_R_inter_prob.nii.gz",
    "SCP_R_superior_prob.nii.gz"]

cp_remote_fnames = [
    "40897772",
    "40897787",
    "40897805",
    "40897769",
    "40897778",
    "40897799",
    "40897784",
    "40897808",
    "40897772",
    "40897781",
    "40897793",
    "40897790",
    "40897775",
    "40897796"]

cp_md5_hashes = [
    "7ec3a78f30aefe8c7e2a99773941b8f4",
    "fc6b088f359201aff480d8e2f7b17eef",
    "e9688ff4554b768f79a64731705107f9",
    "f1959600d6aaae58d7ba4ce5e63c5e3b",
    "e68c669d58bd2cbb3490bd952957009e",
    "eb7643fd54f6046cd71b7ac085679594",
    "746da79d0639630c1eb872b2895814ee",
    "3f69b20a8a7dec4839cef461667d18a5",
    "4d95b3d3804352eae6e66fe024f1baf5",
    "a9410ab5fe3240cc8ef929062d54cc9e",
    "531babda0dd284bda4a5b4b1303f8266",
    "dde4907d7914dfe71ed07436b073bd75",
    "c02f2fcb48c33fe4b2988087075e9566",
    "62e32e090cd326790c3fdbc6acb8eb75"]

fetch_cp_templates = _make_reusable_fetcher(
    "fetch_cb_templates",
    op.join(afq_home,
            'cp_templates'),
    baseurl, cp_remote_fnames,
    cp_fnames,
    md5_list=cp_md5_hashes,
    doc="Download AFQ cerebellar penducles templates")


def read_cp_templates(as_img=True, resample_to=False):
    """Load AFQ Cerebellar penducles templates from file

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict with: keys: names of template ROIs and values: nibabel Nifti1Image
    objects from each of the ROI nifti files.
    """
    logger = logging.getLogger('AFQ')

    logger.debug('loading or templates')
    tic = time.perf_counter()

    template_dict = _fetcher_to_template(
        fetch_cp_templates,
        as_img=as_img,
        resample_to=resample_to)

    toc = time.perf_counter()
    logger.debug(
        f'Cerebellar peduncles templates loaded in {toc - tic:0.4f} seconds')

    return template_dict


or_fnames = [
    "left_thal_MNI.nii.gz",
    "left_V1_MNI.nii.gz",
    "right_thal_MNI.nii.gz",
    "right_V1_MNI.nii.gz",
    "left_OP_MNI.nii.gz",
    "left_OR_1.nii.gz",
    "left_OR_2.nii.gz",
    "left_pos_thal_MNI.nii.gz",
    "left_TP_MNI.nii.gz",
    "right_OP_MNI.nii.gz",
    "right_OR_1.nii.gz",
    "right_OR_2.nii.gz",
    "right_pos_thal_MNI.nii.gz",
    "right_TP_MNI.nii.gz",
]

or_remote_fnames = [
    "36514170",
    "26831633",
    "36514173",
    "26831639",
    "26831642",
    "26831645",
    "26831648",
    "26831651",
    "26831654",
    "26831657",
    "26831660",
    "26831663",
    "26831666",
    "26831669",
]

or_md5_hashes = [
    "a45126a727c4b5d843b2f7aae181825f",
    "ad996c67bf5cc59fc3a7b60255873b67",
    "7a75c3ddd25335277a099626dbc946ac",
    "cc88fb4671311404eb9dfa8fa11a59e0",
    "9cff03af586d9dd880750cef3e0bf63f",
    "ff728ba3ffa5d1600bcd19fdef8182c4",
    "4f1978e418a3169609375c28b3eba0fd",
    "fd163893081b520f4594171aeea04f39",
    "bf795d197912b5e074d248d2763c6930",
    "13efde1efe0de52683cbf352ecba457e",
    "75c7bd2092950578e599a2dcb218909f",
    "8f3890fa8c26a568503226757f7e7d6c",
    "f239aa3140809152da8884ff879dde1b",
    "60a748567e4dd81b40ad8967a14cb09e",
]

ar_fnames = [
    "AAL_Thal_R.nii.gz",
    "AAL_Thal_L.nii.gz",
    "AAL_TempSup_L.nii.gz",
    "AAL_TempSup_R.nii.gz",
]

ar_remote_fnames = [
    "43817367",
    "43817370",
    "43817373",
    "43817376",
]

ar_md5_hashes = [
    "5f2c0c0f1b7b32ba0a8474a460a2b980",
    "a8d308a93b26242c04b878c733cb252f",
    "92a255f706dd4e09268c5bc6bf2876e4",
    "ec0a3ef3cb3a4c549dd783dc8aeceeee",
]


fetch_ar_templates = _make_reusable_fetcher(
    "fetch_ar_templates",
    op.join(afq_home,
            'ar_templates'),
    baseurl, ar_remote_fnames,
    ar_fnames,
    md5_list=ar_md5_hashes,
    doc="Download AFQ or templates")


def read_ar_templates(as_img=True, resample_to=False):
    """Load AFQ AR templates from file

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict with: keys: names of template ROIs and values: nibabel Nifti1Image
    objects from each of the ROI nifti files.
    """
    logger = logging.getLogger('AFQ')

    logger.debug('loading ar templates')
    tic = time.perf_counter()

    template_dict = _fetcher_to_template(
        fetch_ar_templates,
        as_img=as_img,
        resample_to=resample_to)

    toc = time.perf_counter()
    logger.debug(f'or templates loaded in {toc - tic:0.4f} seconds')

    return template_dict


fetch_or_templates = _make_reusable_fetcher(
    "fetch_or_templates",
    op.join(afq_home,
            'or_templates'),
    baseurl, or_remote_fnames,
    or_fnames,
    md5_list=or_md5_hashes,
    doc="Download AFQ or templates")


def read_or_templates(as_img=True, resample_to=False):
    """Load AFQ OR templates from file

    Parameters
    ----------
    as_img : bool, optional
        If True, values are `Nifti1Image`. Otherwise, values are
        paths to Nifti files. Default: True
    resample_to : str or nibabel image class instance, optional
        A template image to resample to. Typically, this should be the
        template to which individual-level data are registered. Defaults to
        the MNI template. Default: False

    Returns
    -------
    dict with: keys: names of template ROIs and values: nibabel Nifti1Image
    objects from each of the ROI nifti files.
    """
    logger = logging.getLogger('AFQ')

    logger.debug('loading or templates')
    tic = time.perf_counter()

    template_dict = _fetcher_to_template(
        fetch_or_templates,
        as_img=as_img,
        resample_to=resample_to)

    toc = time.perf_counter()
    logger.debug(f'or templates loaded in {toc - tic:0.4f} seconds')

    return template_dict


stanford_hardi_tractography_remote_fnames = ["5325715", "5325718", "25289735"]
stanford_hardi_tractography_hashes = ['6f4bdae702031a48d1cd3811e7a42ef9',
                                      'f20854b4f710577c58bd01072cfb4de6',
                                      '294bfd1831861e8635eef8834ff18892']
stanford_hardi_tractography_fnames = [
    'mapping.nii.gz',
    'tractography_subsampled.trk',
    'full_segmented_cleaned_tractography.trk']

fetch_stanford_hardi_tractography = _make_reusable_fetcher(
    "fetch_stanford_hardi_tractography",
    op.join(afq_home,
            'stanford_hardi_tractography'),
    baseurl,
    stanford_hardi_tractography_remote_fnames,
    stanford_hardi_tractography_fnames,
    md5_list=stanford_hardi_tractography_hashes,
    doc="""Download Stanford HARDI tractography and mapping. For testing
           purposes""")


def read_stanford_hardi_tractography():
    """
    Reads a minimal tractography from the Stanford dataset.
    """
    files, folder = fetch_stanford_hardi_tractography()
    files_dict = {}
    files_dict['mapping.nii.gz'] = nib.load(
        op.join(afq_home,
                'stanford_hardi_tractography',
                'mapping.nii.gz'))

    # We need the original data as reference
    dwi_img, gtab = dpd.read_stanford_hardi()

    files_dict['tractography_subsampled.trk'] = load_tractogram(
        op.join(afq_home,
                'stanford_hardi_tractography',
                'tractography_subsampled.trk'),
        dwi_img,
        bbox_valid_check=False,
        trk_header_check=False).streamlines

    files_dict['full_segmented_cleaned_tractography.trk'] = load_tractogram(
        op.join(
            afq_home,
            'stanford_hardi_tractography',
            'full_segmented_cleaned_tractography.trk'),
        dwi_img).streamlines

    return files_dict


def to_bids_description(path, fname='dataset_description.json',
                        BIDSVersion="1.4.0", **kwargs):
    """Dumps a dict into a bids description at the given location"""
    kwargs.update({"BIDSVersion": BIDSVersion})
    desc_file = op.join(path, fname)
    with open(desc_file, 'w') as outfile:
        json.dump(kwargs, outfile)


def organize_cfin_data(path=None):
    """
    Create the expected file-system structure for the
    CFIN multi b-value diffusion data-set.
    """

    dpd.fetch_cfin_multib()
    if path is None:
        os.makedirs(afq_home, exist_ok=True)
        path = afq_home

    bids_path = op.join(path, 'cfin_multib',)
    derivatives_path = op.join(bids_path, 'derivatives')
    dmriprep_folder = op.join(derivatives_path, 'dmriprep')

    if not op.exists(derivatives_path):
        anat_folder = op.join(dmriprep_folder, 'sub-01', 'ses-01', 'anat')
        os.makedirs(anat_folder, exist_ok=True)
        dwi_folder = op.join(dmriprep_folder, 'sub-01', 'ses-01', 'dwi')
        os.makedirs(dwi_folder, exist_ok=True)
        t1_img = dpd.read_cfin_t1()
        nib.save(t1_img, op.join(anat_folder, 'sub-01_ses-01_T1w.nii.gz'))
        dwi_img, gtab = dpd.read_cfin_dwi()
        nib.save(dwi_img, op.join(dwi_folder, 'sub-01_ses-01_dwi.nii.gz'))
        np.savetxt(op.join(dwi_folder, 'sub-01_ses-01_dwi.bvec'), gtab.bvecs)
        np.savetxt(op.join(dwi_folder, 'sub-01_ses-01_dwi.bval'), gtab.bvals)

    to_bids_description(
        bids_path,
        **{"BIDSVersion": "1.0.0",
           "Name": "CFIN",
           "Subjects": ["sub-01"]})

    to_bids_description(
        dmriprep_folder,
        **{"Name": "CFIN",
           "PipelineDescription": {"Name": "dipy"},
           "GeneratedBy": [{"Name": "dipy"}]})


def organize_stanford_data(path=None, clear_previous_afq=None):
    """
    If necessary, downloads the Stanford HARDI dataset into DIPY directory and
    creates a BIDS compliant file-system structure in AFQ data directory:


    ~/AFQ_data/
    └── stanford_hardi
    ├── dataset_description.json
    └── derivatives
        ├── freesurfer
        │   ├── dataset_description.json
        │   └── sub-01
        │       └── ses-01
        │           └── anat
        │               ├── sub-01_ses-01_T1w.nii.gz
        │               └── sub-01_ses-01_seg.nii.gz
        └── vistasoft
            ├── dataset_description.json
            └── sub-01
                └── ses-01
                    └── dwi
                        ├── sub-01_ses-01_dwi.bval
                        ├── sub-01_ses-01_dwi.bvec
                        └── sub-01_ses-01_dwi.nii.gz

    Parameters
    ----------
    path : str or None
        Path to download dataset to, by default it is ~/AFQ_data/.
    clear_previous_afq : str or None
        Whether to clear previous afq results in the stanford
        hardi dataset. If not None, can be "all", "track", "recog", "prof".
        Default: None
    """
    logger = logging.getLogger('AFQ')

    # fetches data for first subject and session
    logger.info('fetching Stanford HARDI data')
    dpd.fetch_stanford_hardi()

    if path is None:
        if not op.exists(afq_home):
            logger.info(f'creating AFQ home directory: {afq_home}')
        os.makedirs(afq_home, exist_ok=True)
        path = afq_home

    bids_path = op.join(path, 'stanford_hardi',)
    derivatives_path = op.join(bids_path, 'derivatives')
    dmriprep_folder = op.join(derivatives_path, 'vistasoft')
    freesurfer_folder = op.join(derivatives_path, 'freesurfer')

    if clear_previous_afq is not None and op.exists(derivatives_path):
        afq_folder = op.join(derivatives_path, 'afq')
        if clear_previous_afq == "all":
            if op.exists(afq_folder):
                shutil.rmtree(afq_folder)
        else:
            apply_cmd_to_afq_derivs(
                op.join(afq_folder, "sub-01/ses-01"),
                op.join(afq_folder, "sub-01/ses-01/sub-01_ses-01"),
                dependent_on=clear_previous_afq)

    if not op.exists(derivatives_path):
        logger.info(f'creating derivatives directory: {derivatives_path}')

        # anatomical data
        anat_folder = op.join(freesurfer_folder, 'sub-01', 'ses-01', 'anat')
        os.makedirs(anat_folder, exist_ok=True)

        t1_img = dpd.read_stanford_t1()
        nib.save(t1_img, op.join(anat_folder, 'sub-01_ses-01_T1w.nii.gz'))

        seg_img = dpd.read_stanford_labels()[-1]
        nib.save(seg_img, op.join(anat_folder,
                                  'sub-01_ses-01_seg.nii.gz'))

        # diffusion-weighted imaging data
        dwi_folder = op.join(dmriprep_folder, 'sub-01', 'ses-01', 'dwi')
        os.makedirs(dwi_folder, exist_ok=True)

        dwi_img, gtab = dpd.read_stanford_hardi()
        nib.save(dwi_img, op.join(dwi_folder, 'sub-01_ses-01_dwi.nii.gz'))
        np.savetxt(op.join(dwi_folder, 'sub-01_ses-01_dwi.bvec'), gtab.bvecs)
        np.savetxt(op.join(dwi_folder, 'sub-01_ses-01_dwi.bval'), gtab.bvals)
    else:
        logger.info('Dataset is already in place. If you want to fetch it '
                    + 'again please first remove the folder '
                    + derivatives_path)

    # Dump out the description of the dataset
    to_bids_description(bids_path,
                        **{"Name": "Stanford HARDI", "Subjects": ["sub-01"]})

    # And descriptions of the pipelines in the derivatives:
    to_bids_description(dmriprep_folder,
                        **{"Name": "Stanford HARDI",
                           "PipelineDescription": {"Name": "vistasoft"},
                           "GeneratedBy": [{"Name": "vistasoft"}]})
    to_bids_description(freesurfer_folder,
                        **{"Name": "Stanford HARDI",
                           "PipelineDescription": {"Name": "freesurfer"},
                           "GeneratedBy": [{"Name": "freesurfer"}]})


fetch_stanford_hardi_lv1 = _make_reusable_fetcher(
    "fetch_stanford_hardi_lv1",
    op.join(afq_home,
            'stanford_hardi',
            'derivatives/freesurfer/sub-01/ses-01/anat'),
    'https://stacks.stanford.edu/file/druid:ng782rw8378/',
    ["SUB1_LV1.nii.gz"],
    ["sub-01_ses-01_desc-LV1_anat.nii.gz"],
    md5_list=["e403c602e53e5491414f86af5f08a913"],
    doc="Download the LV1 segmentation for the Standord Hardi subject",
    unzip=False)


fetch_hcp_atlas_16_bundles = _make_reusable_fetcher(
    "fetch_hcp_atlas_16_bundles",
    op.join(afq_home,
            'hcp_atlas_16_bundles'),
    'https://ndownloader.figshare.com/files/',
    ["11921522"],
    ["atlas_16_bundles.zip"],
    md5_list=["b071f3e851f21ba1749c02fc6beb3118"],
    doc="Download minimal Recobundles atlas",
    unzip=True)


fetch_hcp_atlas_80_bundles = _make_reusable_fetcher(
    "fetch_hcp_atlas_80_bundles",
    op.join(afq_home,
            'hcp_atlas_80_bundles'),
    'https://ndownloader.figshare.com/files/',
    ["13638644"],
    ["Atlas_80_Bundles.zip"],
    md5_list=["78331d527a10ec000d4f33bac472e099"],
    doc="Download 80-bundle Recobundles atlas",
    unzip=True)


def read_hcp_atlas(n_bundles=16, as_file=False):
    """
    as_file : bool, optional
        If True, values are paths to sls. Otherwise, the sl
        are located and the centroids calculated. Default: False

    n_bundles : int
        16 or 80, which selects among the two different
        atlases:

        https://figshare.com/articles/Simple_model_bundle_atlas_for_RecoBundles/6483614  # noqa

        https://figshare.com/articles/Advanced_Atlas_of_80_Bundles_in_MNI_space/7375883  # noqa
    """
    bundle_dict = {}
    if n_bundles == 16:
        _, folder = fetch_hcp_atlas_16_bundles()
        atlas_folder = "Atlas_in_MNI_Space_16_bundles"
    elif n_bundles == 80:
        _, folder = fetch_hcp_atlas_80_bundles()
        atlas_folder = "Atlas_80_Bundles"

    bundle_files = glob(
        op.join(
            folder,
            atlas_folder,
            "bundles", "*.trk"))
    centroid_folder = op.join(
        folder,
        atlas_folder,
        "centroid")
    os.makedirs(centroid_folder, exist_ok=True)
    for bundle_file in bundle_files:
        bundle = drop_extension(op.split(bundle_file)[-1])
        centroid_file = op.join(centroid_folder, f"{bundle}.trk")
        bundle_dict[bundle] = {"recobundles": {}}
        if not op.exists(centroid_file):
            bundle_sl = load_tractogram(
                bundle_file,
                'same',
                bbox_valid_check=False)
            feature = ResampleFeature(nb_points=100)
            metric = AveragePointwiseEuclideanMetric(feature)
            qb = QuickBundles(np.inf, metric=metric)
            cluster = [qb.cluster(bundle_sl.streamlines).centroids[0]]
            save_tractogram(
                StatefulTractogram(
                    cluster, bundle_sl, Space.RASMM),
                centroid_file,
                bbox_valid_check=False)
        if not as_file:
            bundle_dict[bundle]["recobundles"]['sl'] = load_tractogram(
                bundle_file,
                'same',
                bbox_valid_check=False).streamlines
            bundle_dict[bundle]["recobundles"]['centroid'] = load_tractogram(
                centroid_file,
                "same", bbox_valid_check=False).streamlines
        else:
            bundle_dict[bundle]["recobundles"]['sl'] = bundle_file
            bundle_dict[bundle]["recobundles"]['centroid'] = centroid_file

    # For some reason, this file-name has a 0 in it, instead of an O:
    bundle_dict["IFOF_R"] = bundle_dict["IF0F_R"]
    # In the 80-bundle case, there are two files, and both have identical
    # content, so this is fine:
    del bundle_dict["IF0F_R"]
    return bundle_dict


fetch_aal_atlas = _make_reusable_fetcher(
    "fetch_aal_atlas",
    op.join(afq_home,
            'aal_atlas'),
    'https://ndownloader.figshare.com/files/',
    ["28416852",
     "28416855"],
    ["MNI_AAL_AndMore.nii.gz",
     "MNI_AAL.txt"],
    md5_list=["69395b75a16f00294a80eb9428bf7855",
              "59fd3284b17de2fbe411ca1c7afe8c65"],
    doc="Download the AAL atlas",
    unzip=False)


def read_aal_atlas(resample_to=None):
    """
    Reads the AAL atlas [1]_.

    Parameters
    ----------
    template : nib.Nifti1Image class instance, optional
        If provided, this is the template used and AAL atlas should be
        registered and aligned to this template


    .. [1] Tzourio-Mazoyer N, Landeau B, Papathanassiou D, Crivello F, Etard O,
           Delcroix N, Mazoyer B, Joliot M. (2002). Automated anatomical
           labeling of activations in SPM using a macroscopic anatomical
           parcellation of the MNI MRI single-subject brain. Neuroimage. 2002;
           15(1):273-89.
    """
    file_dict, folder = fetch_aal_atlas()
    out_dict = {}
    for f in file_dict:
        if f.endswith('.txt'):
            out_dict['labels'] = pd.read_csv(op.join(folder, f))
        else:
            out_dict['atlas'] = nib.load(op.join(folder, f))
    if resample_to is not None:
        data = out_dict['atlas'].get_fdata()
        oo = []
        for ii in range(data.shape[-1]):
            oo.append(resample(
                data[..., ii],
                resample_to,
                out_dict['atlas'].affine,
                resample_to.affine).get_fdata())
        out_dict['atlas'] = nib.Nifti1Image(np.stack(oo, -1),
                                            resample_to.affine)
    return out_dict


def _apply_mask(template_img, resolution=1):
    """
    Helper function, gets MNI brain mask and applies it to template_img.

    Parameters
    ----------
    template_img : nib.Nifti1Image
        Unmasked template
    resolution : int, optional
        Resolution of mask. Default: 1

    Returns
    -------
    Masked template as nib.Nifti1Image
    """
    mask_img = nib.load(str(tflow.get('MNI152NLin2009cAsym',
                                      resolution=resolution,
                                      desc='brain',
                                      suffix='mask')))

    template_data = template_img.get_fdata()
    mask_data = mask_img.get_fdata()

    if mask_data.shape != template_data.shape:
        mask_img = nib.Nifti1Image(
            resample(
                mask_data,
                template_data,
                mask_img.affine,
                template_img.affine).get_fdata(),
            template_img.affine)
        mask_data = mask_img.get_fdata()

    out_data = template_data * mask_data
    return nib.Nifti1Image(out_data, template_img.affine)


def read_mni_template(resolution=1, mask=True, weight="T2w"):
    """

    Reads the MNI T1w or T2w template

    Parameters
    ----------
    resolution : int, optional.
        Either 1 or 2, the resolution in mm of the voxels. Default: 1.

    mask : bool, optional
        Whether to mask the data with a brain-mask before returning the image.
        Default : True

    weight: str, optional
        Which relaxation technique to use.
        Should be either "T2w" or "T1w".
        Default : "T2w"

    Returns
    -------
    nib.Nifti1Image class instance
    containing masked or unmasked T1w or template.

    """
    template_img = nib.load(str(tflow.get('MNI152NLin2009cAsym',
                                          desc=None,
                                          resolution=resolution,
                                          suffix=weight,
                                          extension='nii.gz')))
    if not mask:
        return template_img
    else:
        return _apply_mask(template_img, resolution)


fetch_biobank_templates = \
    _make_reusable_fetcher(
        "fetch_biobank_templates",
        op.join(afq_home,
                'biobank_templates'),
        "http://biobank.ctsu.ox.ac.uk/showcase/showcase/docs/",
        ["bmri_group_means.zip"],
        ["bmri_group_means.zip"],
        data_size="1.1 GB",
        doc="Download UK Biobank templates",
        unzip=True)


def read_ukbb_fa_template(mask=True):
    """
    Reads the UK Biobank FA template

    Parameters
    ----------
    mask : bool, optional
        Whether to mask the data with a brain-mask before returning the image.
        Default : True

    Returns
    -------
    nib.Nifti1Image class instance containing the FA template.

    """
    fa_folder = op.join(
        afq_home,
        'biobank_templates',
        'UKBiobank_BrainImaging_GroupMeanTemplates'
    )
    fa_path = op.join(
        fa_folder,
        'dti_FA.nii.gz'
    )

    if not op.exists(fa_path):
        logger = logging.getLogger('AFQ')
        logger.warning(
            "Downloading brain MRI group mean statistics from UK Biobank. "
            + "This download is approximately 1.1 GB. "
            + "It is currently necessary to access the FA template.")

        files, folder = fetch_biobank_templates()

        # remove zip
        for filename in files:
            os.remove(op.join(folder, filename))

        # remove non-FA related directories
        for filename in os.listdir(fa_folder):
            full_path = op.join(fa_folder, filename)
            if full_path != fa_path:
                if os.path.isfile(full_path):
                    os.remove(full_path)
                else:
                    shutil.rmtree(full_path)

    template_img = nib.load(fa_path)

    if not mask:
        return template_img
    else:
        return _apply_mask(template_img, 1)


def fetch_hcp(subjects,
              hcp_bucket='hcp-openaccess',
              profile_name="hcp",
              path=None,
              study='HCP_1200',
              aws_access_key_id=None,
              aws_secret_access_key=None):
    """
    Fetch HCP diffusion data and arrange it in a manner that resembles the
    BIDS [1]_ specification.

    Parameters
    ----------
    subjects : list
        Each item is an integer, identifying one of the HCP subjects
    hcp_bucket : string, optional
        The name of the HCP S3 bucket. Default: "hcp-openaccess"
    profile_name : string, optional
        The name of the AWS profile used for access. Default: "hcp"
    path : string, optional
        Path to save files into. Default: '~/AFQ_data'
    study : string, optional
        Which HCP study to grab. Default: 'HCP_1200'
    aws_access_key_id : string, optional
        AWS credentials to HCP AWS S3. Will only be used if `profile_name` is
        set to False.
    aws_secret_access_key : string, optional
        AWS credentials to HCP AWS S3. Will only be used if `profile_name` is
        set to False.

    Returns
    -------
    dict with remote and local names of these files,
    path to BIDS derivative dataset

    Notes
    -----
    To use this function with its default setting, you need to have a
    file '~/.aws/credentials', that includes a section:

    [hcp]
    AWS_ACCESS_KEY_ID=XXXXXXXXXXXXXXXX
    AWS_SECRET_ACCESS_KEY=XXXXXXXXXXXXXXXX

    The keys are credentials that you can get from HCP
    (see https://wiki.humanconnectome.org/display/PublicData/How+To+Connect+to+Connectome+Data+via+AWS)  # noqa

    Local filenames are changed to match our expected conventions.

    .. [1] Gorgolewski et al. (2016). The brain imaging data structure,
           a format for organizing and describing outputs of neuroimaging
           experiments. Scientific Data, 3::160044. DOI: 10.1038/sdata.2016.44.
    """
    if profile_name:
        boto3.setup_default_session(profile_name=profile_name)
    elif aws_access_key_id is not None and aws_secret_access_key is not None:
        boto3.setup_default_session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key)
    else:
        raise ValueError("Must provide either a `profile_name` or ",
                         "both `aws_access_key_id` and ",
                         "`aws_secret_access_key` as input to 'fetch_hcp'")

    s3 = boto3.resource('s3')
    bucket = s3.Bucket(hcp_bucket)

    if path is None:
        if not op.exists(afq_home):
            os.mkdir(afq_home)
        my_path = afq_home
    else:
        my_path = path

    base_dir = op.join(my_path, study, 'derivatives', 'dmriprep')

    if not os.path.exists(base_dir):
        os.makedirs(base_dir, exist_ok=True)

    data_files = {}
    for subject in subjects:
        # We make a single session folder per subject for this case, because
        # AFQ api expects session structure:
        sub_dir = op.join(base_dir, f'sub-{subject}')
        sess_dir = op.join(sub_dir, "ses-01")
        if not os.path.exists(sub_dir):
            os.makedirs(os.path.join(sess_dir, 'dwi'), exist_ok=True)
            os.makedirs(os.path.join(sess_dir, 'anat'), exist_ok=True)
        data_files[op.join(sess_dir, 'dwi', f'sub-{subject}_dwi.bval')] =\
            f'{study}/{subject}/T1w/Diffusion/bvals'
        data_files[op.join(sess_dir, 'dwi', f'sub-{subject}_dwi.bvec')] =\
            f'{study}/{subject}/T1w/Diffusion/bvecs'
        data_files[op.join(sess_dir, 'dwi', f'sub-{subject}_dwi.nii.gz')] =\
            f'{study}/{subject}/T1w/Diffusion/data.nii.gz'
        data_files[op.join(sess_dir, 'anat', f'sub-{subject}_T1w.nii.gz')] =\
            f'{study}/{subject}/T1w/T1w_acpc_dc.nii.gz'
        data_files[op.join(sess_dir, 'anat',
                           f'sub-{subject}_aparc+aseg_seg.nii.gz')] =\
            f'{study}/{subject}/T1w/aparc+aseg.nii.gz'

    download_files = {}
    for k in data_files.keys():
        if not op.exists(k):
            download_files[k] = data_files[k]
    if len(download_files.keys()):
        with tqdm(total=len(download_files.keys())) as pbar:
            for k in download_files.keys():
                pbar.set_description_str(f"Downloading {k}")
                bucket.download_file(data_files[k], k)
                pbar.update()

    # Create the BIDS dataset description file text
    hcp_acknowledgements = """Data were provided by the Human Connectome Project, WU-Minn Consortium (Principal Investigators: David Van Essen and Kamil Ugurbil; 1U54MH091657) funded by the 16 NIH Institutes and Centers that support the NIH Blueprint for Neuroscience Research; and by the McDonnell Center for Systems Neuroscience at Washington University.""",  # noqa
    to_bids_description(op.join(my_path, study),
                        **{"Name": study,
                           "Acknowledgements": hcp_acknowledgements,
                           "Subjects": subjects})

    # Create the BIDS derivatives description file text
    to_bids_description(base_dir,
                        **{"Name": study,
                           "Acknowledgements": hcp_acknowledgements,
                           "PipelineDescription": {'Name': 'dmriprep'},
                           "GeneratedBy": [{'Name': 'dmriprep'}]})

    return data_files, op.join(my_path, study)


def fetch_hbn_preproc(subjects, path=None):
    """
    Fetches data from the Healthy Brain Network POD2 study [1, 2]_.

    Parameters
    ----------
    subjects : list
        Identifiers of the subjects to download.
        For example: ["NDARAA112DMH", "NDARAA117NEJ"].
    path : string, optional
        Path to save files into. Default: '~/AFQ_data'

    Returns
    -------
    dict with remote and local names of these files,
    path to BIDS derivative dataset

    Notes
    -----

    .. [1] Alexander LM, Escalera J, Ai L, et al. An open resource for
        transdiagnostic research in pediatric mental health and learning
        disorders. Sci Data. 2017;4:170181.

    .. [2] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and
        quality controlled resource for pediatric brain white-matter research.
        Scientific Data. 2022;9(1):1-27.

    """
    # Anonymous access:
    client = boto3.client('s3', config=Config(signature_version=UNSIGNED))

    if path is None:
        if not op.exists(afq_home):
            os.mkdir(afq_home)
        my_path = afq_home
    else:
        my_path = path

    base_dir = op.join(my_path, "HBN", 'derivatives', 'qsiprep')

    if not os.path.exists(base_dir):
        os.makedirs(base_dir, exist_ok=True)

    data_files = {}

    for subject in subjects:
        initial_query = client.list_objects(
            Bucket="fcp-indi",
            Prefix=f"data/Projects/HBN/BIDS_curated/sub-{subject}/")
        ses = initial_query['Contents'][0]["Key"].split('/')[5]
        query = client.list_objects(
            Bucket="fcp-indi",
            Prefix=f"data/Projects/HBN/BIDS_curated/derivatives/qsiprep/sub-{subject}/")  # noqa
        file_list = [kk["Key"] for kk in query["Contents"]]
        sub_dir = op.join(base_dir, f'sub-{subject}')
        ses_dir = op.join(sub_dir, ses)
        if not os.path.exists(sub_dir):
            os.makedirs(os.path.join(sub_dir, 'anat'), exist_ok=True)
            os.makedirs(os.path.join(sub_dir, 'figures'), exist_ok=True)
            os.makedirs(os.path.join(ses_dir, 'dwi'), exist_ok=True)
            os.makedirs(os.path.join(ses_dir, 'anat'), exist_ok=True)
        for remote in file_list:
            full = remote.split("Projects")[-1][1:].replace("/BIDS_curated", "")
            local = op.join(my_path, full)
            data_files[local] = remote

    download_files = {}
    for k in data_files.keys():
        if not op.exists(k):
            download_files[k] = data_files[k]
    if len(download_files.keys()):
        with tqdm(total=len(download_files.keys())) as pbar:
            for k in download_files.keys():
                pbar.set_description_str(f"Downloading {k}")
                client.download_file("fcp-indi", download_files[k], k)
                pbar.update()

    # Create the BIDS dataset description file text
    hbn_acknowledgements = """XXX""",  # noqa
    to_bids_description(op.join(my_path, "HBN"),
                        **{"Name": "HBN",
                           "Acknowledgements": hbn_acknowledgements,
                           "Subjects": subjects})

    # Create the BIDS derivatives description file text
    to_bids_description(base_dir,
                        **{"Name": "HBN",
                           "Acknowledgements": hbn_acknowledgements,
                           "PipelineDescription": {'Name': 'qsiprep'},
                           "GeneratedBy": [{'Name': 'qsiprep'}]})

    return data_files, op.join(my_path, "HBN")


def fetch_hbn_afq(subjects, path=None):
    """
    Fetches AFQ derivatives for Healthy Brain Network POD2 study [1, 2]_.

    Parameters
    ----------
    subjects : list
        Identifiers of the subjects to download.
        For example: ["NDARAA112DMH", "NDARAA117NEJ"].
    path : string, optional
        Path to save files into. Default: '~/AFQ_data'

    Returns
    -------
    dict with remote and local names of these files,
    path to BIDS derivative dataset

    Notes
    -----

    .. [1] Alexander LM, Escalera J, Ai L, et al. An open resource for
        transdiagnostic research in pediatric mental health and learning
        disorders. Sci Data. 2017;4:170181.

    .. [2] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and
        quality controlled resource for pediatric brain white-matter research.
        Scientific Data. 2022;9(1):1-27.

    """
    # Anonymous access:
    client = boto3.client('s3', config=Config(signature_version=UNSIGNED))

    if path is None:
        if not op.exists(afq_home):
            os.mkdir(afq_home)
        my_path = afq_home
    else:
        my_path = path

    base_dir = op.join(my_path, "HBN", 'derivatives', 'afq')

    if not os.path.exists(base_dir):
        os.makedirs(base_dir, exist_ok=True)

    data_files = {}

    for subject in subjects:
        initial_query = client.list_objects(
            Bucket="fcp-indi",
            Prefix=f"data/Projects/HBN/BIDS_curated/sub-{subject}/")
        ses = initial_query['Contents'][0]["Key"].split('/')[5]
        query = client.list_objects(
            Bucket="fcp-indi",
            Prefix=f"data/Projects/HBN/BIDS_curated/derivatives/afq/sub-{subject}/")  # noqa
        file_list = [kk["Key"] for kk in query["Contents"]]
        sub_dir = op.join(base_dir, f'sub-{subject}')
        ses_dir = op.join(sub_dir, ses)

        for deriv_dir in ["bundles",
                          "clean_bundles",
                          "ROIs",
                          "tract_profile_plots",
                          "viz_bundles"]:
            this_deriv = os.path.join(ses_dir, deriv_dir)
            if not os.path.exists(this_deriv):
                os.makedirs(this_deriv, exist_ok=True)
        for remote in file_list:
            full = remote.split("Projects")[-1][1:].replace("/BIDS_curated", "")
            local = op.join(afq_home, full)
            data_files[local] = remote

    download_files = {}
    for k in data_files.keys():
        if not op.exists(k):
            download_files[k] = data_files[k]
    if len(download_files.keys()):
        with tqdm(total=len(download_files.keys())) as pbar:
            for k in download_files.keys():
                pbar.set_description_str(f"Downloading {k}")
                client.download_file("fcp-indi", download_files[k], k)
                pbar.update()

    # Create the BIDS dataset description file text
    hbn_acknowledgements = """XXX""",  # noqa
    to_bids_description(op.join(my_path, "HBN"),
                        **{"Name": "HBN",
                           "Acknowledgements": hbn_acknowledgements,
                           "Subjects": subjects})

    # Create the BIDS derivatives description file text
    to_bids_description(base_dir,
                        **{"Name": "HBN",
                           "PipelineDescription": {'Name': 'afq'},
                           "GeneratedBy": [{'Name': 'afq'}]})

    return data_files, op.join(my_path, "HBN")



"""
====================================
How to segment out only some bundles
====================================

The pyAFQ software can be configured to find all of its default set of white
matter pathways, or bundles. Alternatively, it can be configured to find only
some bundles. This example shows how to track and recognize only certain
bundles that you are interested in, though note that ARC, pARC and VOF are all
also part of the set of bundles that are segmented per default.

"""

import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd
from AFQ.api.group import GroupAFQ
from AFQ.definitions.image import RoiImage
import AFQ.utils.streamlines as aus

import os.path as op

afd.organize_stanford_data(clear_previous_afq="track")

bundle_names = [
    "Left Arcuate", "Right Arcuate",
    "Left Posterior Arcuate", "Right Posterior Arcuate",
    "Left Vertical Occipital", "Right Vertical Occipital"]
bundle_dict = abd.default18_bd()[bundle_names]


myafq = GroupAFQ(
    op.join(afd.afq_home, 'stanford_hardi'),
    bundle_info=bundle_dict,
    preproc_pipeline='vistasoft',
    tracking_params={
        "n_seeds": 50000,
        "random_seeds": True,
        "seed_mask": RoiImage(use_waypoints=True, use_endpoints=True),
    }
)

for b_name in bundle_names:
    b_len = len(aus.SegmentedSFT.fromfile(myafq.export("bundles")[
        "01"]).get_bundle(b_name))

    if b_len < 1:
        raise ValueError(f"{b_name} not found")

myafq.export("all_bundles_figure")["01"][0]



from AFQ.utils.path import drop_extension


def test_drop_extension():
    assert "/my/.example/.path/to_file" == drop_extension(
        "/my/.example/.path/to_file.nii.gz")

    assert "C:\\Documents\\Newsletters\\Summer2018" == drop_extension(
        "C:\\Documents\\Newsletters\\Summer2018.pdf.gz")



import pytest

from AFQ.viz.utils import Viz


def test_viz_name_errors():
    Viz("fury")

    with pytest.raises(
        TypeError,
        match="Visualization backend contain"
        + " either 'plotly' or 'fury'. "
            + "It is currently set to plotli"):
        Viz("plotli")



import numpy as np
import numpy.testing as npt
import pytest

import nibabel as nib

from dipy.io.stateful_tractogram import Space
from dipy.io.streamline import StatefulTractogram

import AFQ.utils.volume as afv
import AFQ.data.fetch as afd


def test_patch_up_roi():
    roi_bad = np.zeros((10, 10, 10))
    roi_good = np.ones((10, 10, 10))

    afv.patch_up_roi(roi_good)
    with pytest.raises(ValueError):
        afv.patch_up_roi(roi_bad)


def test_density_map():
    file_dict = afd.read_stanford_hardi_tractography()

    # subsample even more
    subsampled_tractography = file_dict["tractography_subsampled.trk"][441:444]
    sft = StatefulTractogram(
        subsampled_tractography,
        file_dict["mapping.nii.gz"],
        Space.VOX)
    density_map = afv.density_map(sft)
    npt.assert_equal(int(np.sum(density_map.get_fdata())), 69)

    density_map = afv.density_map(sft, normalize=True)
    npt.assert_equal(density_map.get_fdata().max(), 1)


def test_dice_coeff():
    affine = np.asarray([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 0]])
    img1 = nib.Nifti1Image(
        np.asarray([
            [0.8, 0.9, 0],
            [0, 0, 0],
            [0, 0, 0]]),
        affine)
    img2 = nib.Nifti1Image(
        np.asarray([
            [0.5, 0, 0],
            [0.6, 0, 0],
            [0, 0, 0]]),
        affine)
    npt.assert_equal(afv.dice_coeff(img1, img2), (0.5+0.8)/(0.5+0.6+0.8+0.9))



import tempfile
import enum
import logging

import numpy as np
import pandas as pd

import AFQ.viz.utils as vut

from dipy.tracking.streamline import set_number_of_points

try:
    import plotly
    import plotly.graph_objs as go
    import plotly.io as pio
    from plotly.subplots import make_subplots
    import plotly.express as px
    from plotly.colors import hex_to_rgb
except (ImportError, ModuleNotFoundError):
    raise ImportError(vut.viz_import_msg_error("plotly"))


scope = pio.kaleido.scope
viz_logger = logging.getLogger("AFQ")


def _inline_interact(figure, show, show_inline):
    """
    Helper function to reuse across viz functions
    """
    if show:
        viz_logger.info("Creating interactive figure in HTML file...")
        plotly.offline.plot(figure)

    if show_inline:
        viz_logger.info("Creating interactive figure inline...")
        plotly.offline.init_notebook_mode()
        plotly.offline.iplot(figure)

    return figure


def _to_color_range(num):
    if num < 0:
        num = 0
    if num >= 0.999:
        num = 0.999
    if num <= 0.001:
        num = 0.001
    return num


def _color_arr2str(color_arr, opacity=1.0):
    return (
        f"rgba({_to_color_range(color_arr[0])}, "
        f"{_to_color_range(color_arr[1])}, "
        f"{_to_color_range(color_arr[2])}, "
        f"{_to_color_range(opacity)})"
    )


def set_layout(figure, color=None):
    if color is None:
        color = "rgba(0,0,0,0)"

    figure.update_layout(
        plot_bgcolor=color,
        scene1=dict(
            xaxis=dict(
                showbackground=False, showticklabels=False, title=''),
            yaxis=dict(
                showbackground=False, showticklabels=False, title=''),
            zaxis=dict(
                showbackground=False, showticklabels=False, title=''),
            aspectmode='data'
        )
    )


def _draw_streamlines(figure, sls, dimensions, color, name, cbv=None, cbs=None,
                      sbv_lims=[None, None], flip_axes=[False, False, False],
                      opacity=1.0):
    color = np.asarray(color)

    if len(sls._offsets) > 1:
        plotting_shape = (sls._data.shape[0] + sls._offsets.shape[0])
    else:
        plotting_shape = sls._data.shape[0]
    # dtype object so None can be stored
    x_pts = np.zeros(plotting_shape)
    y_pts = np.zeros(plotting_shape)
    z_pts = np.zeros(plotting_shape)

    if cbs is not None:
        cbs = np.asarray(cbs)
        line_color = np.zeros((plotting_shape, cbs.shape[1]))
        color = cbs[0, :]
    elif cbv is not None:
        if sbv_lims[0] is None:
            sbv_lims[0] = 0
        if sbv_lims[1] is None:
            sbv_lims[1] = cbv.max()

        color_constant = (color / color.max())\
            * (1.4 / (sbv_lims[1] - sbv_lims[0])) + sbv_lims[0]
        line_color = np.zeros((plotting_shape, 3))
    else:
        color_constant = color
        line_color = np.zeros((plotting_shape, 3))
    customdata_tp = np.zeros(plotting_shape)
    customdata_nodes = np.zeros(plotting_shape)

    for sl_index, plotting_offset in enumerate(sls._offsets):
        sl_length = sls._lengths[sl_index]
        sl = sls._data[plotting_offset:plotting_offset + sl_length]

        # add sl to lines
        total_offset = plotting_offset + sl_index
        x_pts[total_offset:total_offset + sl_length] = sl[:, 0]
        y_pts[total_offset:total_offset + sl_length] = sl[:, 1]
        z_pts[total_offset:total_offset + sl_length] = sl[:, 2]

        # don't draw between streamlines
        if len(sls._offsets) > 1:
            x_pts[total_offset + sl_length] = np.nan
            y_pts[total_offset + sl_length] = np.nan
            z_pts[total_offset + sl_length] = np.nan

        if cbs is not None:
            color_constant = cbs[sl_index]

        if cbv is not None:
            brightness = cbv[
                sl[:, 0].astype(int),
                sl[:, 1].astype(int),
                sl[:, 2].astype(int)
            ]

            line_color[total_offset:total_offset + sl_length, :] = \
                np.outer(brightness, color_constant)
            customdata_tp[total_offset:total_offset + sl_length] = brightness
        else:
            line_color[total_offset:total_offset + sl_length, :] = \
                color_constant
            customdata_tp[total_offset:total_offset + sl_length] = 1
        customdata_nodes[total_offset:total_offset + sl_length] =\
            np.arange(sl_length)

        if line_color.shape[1] > 3:
            line_color[total_offset:total_offset + sl_length, 3] = \
                color_constant[3]  # dont shade alpha values

        if len(sls._offsets) > 1:
            line_color[total_offset + sl_length, :] = 0

    if flip_axes[0]:
        x_pts = dimensions[0] - x_pts
    if flip_axes[1]:
        y_pts = dimensions[1] - y_pts
    if flip_axes[2]:
        z_pts = dimensions[2] - z_pts
    hovertext = [
        f'TP: {i1}<br>Node ID: {i2}'
        for i1, i2 in zip(
            customdata_tp, customdata_nodes)]
    figure.add_trace(
        go.Scatter3d(
            x=x_pts,
            y=y_pts,
            z=z_pts,
            name=vut.display_string(name),
            legendgroup=vut.display_string(name),
            marker=dict(
                size=0.0001,
                color=_color_arr2str(color)
            ),  # this is necessary to add color to legend
            line=dict(
                width=8,
                color=line_color,
            ),
            hovertext=hovertext,
            hoverinfo='all',
            opacity=opacity
        ),
        row=1, col=1
    )
    return color_constant


def _plot_profiles(profiles, bundle_name, color, fig, scalar):
    if isinstance(profiles, pd.DataFrame):
        profiles = profiles[profiles.tractID == bundle_name]
        x = profiles["nodeID"]
        y = profiles[scalar]
        line_color = []
        for scalar_val in profiles[scalar].to_numpy():
            line_color.append(_color_arr2str(scalar_val * color))
    else:
        x = np.arange(len(profiles))
        y = profiles
        line_color = []
        for indiv_color in color:
            line_color.append(_color_arr2str(indiv_color))

    fig.add_trace(
        go.Scatter3d(
            x=x,
            y=y,
            z=np.zeros(len(y)),
            name=vut.display_string(bundle_name),
            line=dict(color=line_color, width=15),
            mode="lines",
            legendgroup=vut.display_string(bundle_name)),
        row=1, col=2)

    font = dict(size=20, family="Overpass")
    fixed_camera_for_2d = dict(
        projection=dict(type="orthographic"),
        up=dict(x=0, y=1, z=0),
        eye=dict(x=0, y=0, z=1),
        center=dict(x=0, y=0, z=0))
    fig.update_layout(
        margin={"t": 15, "b": 0, "l": 0, "r": 0},
        scene2=dict(
            camera=fixed_camera_for_2d,
            zaxis=dict(visible=False),
            dragmode=False,
            xaxis_title=dict(text="Location", font=font),
            yaxis_title=dict(text=vut.display_string(scalar), font=font)))


def visualize_bundles(sft, n_points=None,
                      bundle=None, colors=None, shade_by_volume=None,
                      color_by_streamline=None, n_sls_viz=3600,
                      sbv_lims=[None, None], include_profiles=(None, None),
                      flip_axes=[False, False, False], opacity=1.0,
                      figure=None, background=(1, 1, 1), interact=False,
                      inline=False, **kwargs):
    """
    Visualize bundles in 3D

    Parameters
    ----------
    sft : Stateful Tractogram, str
        A Stateful Tractogram containing streamline information
        or a path to a trk file.
        In order to visualize individual bundles, the Stateful Tractogram
        must contain a bundle key in it's data_per_streamline which is a list
        of bundle `'uid'`.

    n_points : int or None
        n_points to resample streamlines to before plotting. If None, no
        resampling is done.

    bundle : str, optional
        The name of a bundle to select
        or an integer for selection from the sft metadata.

    colors : dict or list
        If this is a dict, keys are bundle names and values are RGB tuples.
        If this is a list, each item is an RGB tuple. Defaults to a dict
        from bundles to Tableau 20 RGB values.

    shade_by_volume : ndarray or str, optional
        3d volume use to shade the bundles. If None, no shading
        is performed. Only works when using the plotly backend.
        Default: None

    color_by_streamline : ndarray or dict, optional
        N by 3 array, where N is the number of streamlines in sft;
        for each streamline you specify 3 values between 0 and 1 for rgb.
        If sft has multiple bundles, then use a dict for color_by_streamline,
        where keys are bundle names and values are n by 3 arrays.
        Overrides colors for bundles in the keys
        of the dict if passing a  dict, or for all streamlines if using
        ndarray.
        Default: None

    n_sls_viz : int
        Number of streamlines to randomly select if plotting
        all bundles. Selections will be proportional to the original number of
        streamlines per bundle.
        Default: 3600

    sbv_lims : ndarray
        Of the form (lower bound, upper bound). Shading based on
        shade_by_volume will only differentiate values within these bounds.
        If lower bound is None, will default to 0.
        If upper bound is None, will default to the maximum value in
        shade_by_volume.
        Default: [None, None]

    include_profiles : Tuple of Pandas Dataframe and string
        The first element of the uple is a
        Pandas Dataframe containing profiles in the standard pyAFQ
        output format for the bundle(s) being displayed. It will be used
        to generate a graph of the tract profiles for each bundle,
        with colors corresponding to the colors on the bundles. The string
        is the scalar to use from the profile. If these are None,
        no tract profiles will be graphed.
        Defualt: (None, None)

    flip_axes : ndarray
        Which axes to flip, to orient the image as RAS, which is how we
        visualize.
        For example, if the input image is LAS, use [True, False, False].
        Default: [False, False, False]

    opacity : float
        Float between 0 and 1 defining the opacity of the bundle.
        Default: 1.0

    background : tuple, optional
        RGB values for the background. Default: (1, 1, 1), which is white
        background.

    figure : Plotly Figure object, optional
        If provided, the visualization will be added to this Figure. Default:
        Initialize a new Figure.

    interact : bool
        Whether to open the visualization in an interactive window.
        Default: False

    inline : bool
        Whether to embed the interactivevisualization inline in a notebook.
        Only works in the notebook context. Default: False.

    Returns
    -------
    Plotly Figure object
    """

    if shade_by_volume is not None:
        shade_by_volume = vut.load_volume(shade_by_volume)

    if figure is None:
        if include_profiles[0] is None:
            figure = make_subplots(
                rows=1, cols=1,
                specs=[[{"type": "scene"}]])
        else:
            figure = make_subplots(
                rows=1, cols=2,
                specs=[[{"type": "scene"}, {"type": "scene"}]])

    set_layout(figure, color=_color_arr2str(background))

    for (sls, color, name, dimensions) in vut.tract_generator(
            sft, bundle, colors, n_points,
            n_sls_viz=n_sls_viz):
        if isinstance(color_by_streamline, dict):
            if name in color_by_streamline:
                cbs = color_by_streamline[name]
        else:
            cbs = color_by_streamline

        color_constant = _draw_streamlines(
            figure,
            sls,
            dimensions,
            color,
            name,
            cbv=shade_by_volume,
            cbs=cbs,
            sbv_lims=sbv_lims,
            flip_axes=flip_axes,
            opacity=opacity)
        if include_profiles[0] is not None:
            _plot_profiles(
                include_profiles[0], name, color_constant,
                figure, include_profiles[1])

    figure.update_layout(legend=dict(itemsizing="constant"))
    return _inline_interact(figure, interact, inline)


def create_gif(figure,
               file_name,
               n_frames=30,
               zoom=2.5,
               z_offset=0.5,
               size=(600, 600)):
    """
    Convert a Plotly Figure object into a gif

    Parameters
    ----------
    figure: Plotly Figure object
        Figure to be converted to a gif

    file_name: str
        File to save gif to.

    n_frames: int, optional
        Number of frames in gif.
        Will be evenly distributed throughout the rotation.
        Default: 60

    zoom: float, optional
        How much to magnify the figure in the fig.
        Default: 2.5

    size: tuple, optional
        Size of the gif.
        Default: (600, 600)
    """
    tdir = tempfile.gettempdir()

    for i in range(n_frames):
        theta = (i * 6.28) / n_frames
        camera = dict(
            eye=dict(x=np.cos(theta) * zoom,
                     y=np.sin(theta) * zoom, z=z_offset)
        )
        figure.update_layout(scene_camera=camera)
        figure.write_image(tdir + f"/tgif{i}.png")
        scope._shutdown_kaleido()  # temporary fix for memory leak

    vut.gif_from_pngs(tdir, file_name, n_frames,
                      png_fname="tgif", add_zeros=False)


def _draw_roi(figure, roi, name, color, opacity, dimensions, flip_axes):
    roi = np.where(roi == 1)
    pts = []
    for i, flip in enumerate(flip_axes):
        if flip:
            pts.append(dimensions[i] - (roi[i] + 1))
        else:
            pts.append(roi[i] + 1)
    figure.add_trace(
        go.Scatter3d(
            x=pts[0],
            y=pts[1],
            z=pts[2],
            name=name,
            marker=dict(color=_color_arr2str(color, opacity=opacity)),
            line=dict(color="rgba(0,0,0,0)")
        ),
        row=1, col=1
    )


def visualize_roi(roi, affine_or_mapping=None, static_img=None,
                  roi_affine=None, static_affine=None, reg_template=None,
                  name='ROI', figure=None, flip_axes=[False, False, False],
                  color=np.array([0.9999, 0, 0]),
                  opacity=1.0, interact=False, inline=False):
    """
    Render a region of interest into a volume

    Parameters
    ----------
    roi : str or Nifti1Image
        The ROI information

    affine_or_mapping : ndarray, Nifti1Image, or str, optional
       An affine transformation or mapping to apply to the ROIs before
       visualization. Default: no transform.

    static_img: str or Nifti1Image, optional
        Template to resample roi to.
        Default: None

    roi_affine: ndarray, optional
        Default: None

    static_affine: ndarray, optional
        Default: None

    reg_template: str or Nifti1Image, optional
        Template to use for registration.
        Default: None

    name: str, optional
        Name of ROI for the legend.
        Default: 'ROI'

    color : ndarray, optional
        RGB color for ROI.
        Default: np.array([0.9999, 0, 0])

    opacity : float, optional
        Opacity of ROI.
        Default: 1.0

    flip_axes : ndarray
        Which axes to flip, to orient the image as RAS, which is how we
        visualize.
        For example, if the input image is LAS, use [True, False, False].
        Default: [False, False, False]

    figure : Plotly Figure object, optional
        If provided, the visualization will be added to this Figure. Default:
        Initialize a new Figure.

    interact : bool
        Whether to open the visualization in an interactive window.
        Default: False

    inline : bool
        Whether to embed the interactive visualization inline in a notebook.
        Only works in the notebook context. Default: False.

    Returns
    -------
    Plotly Figure object
    """
    roi = vut.prepare_roi(roi, affine_or_mapping, static_img,
                          roi_affine, static_affine, reg_template)

    if figure is None:
        figure = make_subplots(
            rows=1, cols=1,
            specs=[[{"type": "scene"}]])

    set_layout(figure)

    _draw_roi(figure, roi, name, color, opacity, roi.shape, flip_axes)

    return _inline_interact(figure, interact, inline)


class Axes(enum.IntEnum):
    X = 0
    Y = 1
    Z = 2


def _draw_slice(figure, axis, volume, opacity=0.3, pos=0.5,
                colorscale="greys", invert_colorscale=False):
    height = int(volume.shape[axis] * pos)

    v_min = volume.min()
    sf = volume.max() - v_min

    if axis == Axes.X:
        X, Y, Z = np.mgrid[height:height + 1,
                           :volume.shape[1], :volume.shape[2]]
        values = volume[height, :, :].flatten()
    elif axis == Axes.Y:
        X, Y, Z = np.mgrid[:volume.shape[0],
                           height:height + 1, :volume.shape[2]]
        values = volume[:, height, :].flatten()
    elif axis == Axes.Z:
        X, Y, Z = np.mgrid[:volume.shape[0],
                           :volume.shape[1], height:height + 1]
        values = volume[:, :, height].flatten()

    values = (values - v_min) / sf
    if invert_colorscale:
        values = 1 - values

    figure.add_trace(
        go.Volume(
            x=X.flatten(),
            y=Y.flatten(),
            z=Z.flatten(),
            value=values,
            colorscale=colorscale,
            surface_count=1,
            showscale=False,
            opacity=opacity,
            name=_name_from_enum(axis),
            hoverinfo='skip',
            showlegend=True
        ),
        row=1, col=1
    )


def _name_from_enum(axis):
    if axis == Axes.X:
        return "Sagittal"
    elif axis == Axes.Y:
        return "Coronal"
    elif axis == Axes.Z:
        return "Axial"


def visualize_volume(volume, figure=None, x_pos=0.5, y_pos=0.5,
                     z_pos=0.5, interact=False, inline=False, opacity=0.3,
                     colorscale="gray", invert_colorscale=False,
                     flip_axes=[False, False, False]):
    """
    Visualize a volume

    Parameters
    ----------
    volume : ndarray or str
        3d volume to visualize.

    figure : Plotly Figure object, optional
        If provided, the visualization will be added to this Figure. Default:
        Initialize a new Figure.

    x_pos : float or None, optional
        Where to show Coronal Slice. If None, slice is not shown.
        Should be a decimal between 0 and 1.
        Indicatesthe fractional position along the perpendicular
        axis to the slice.
        Default: 0.5

    y_pos : float or None, optional
        Where to show Sagittal Slice. If None, slice is not shown.
        Should be a decimal between 0 and 1.
        Indicatesthe fractional position along the perpendicular
        axis to the slice.
        Default: 0.5

    z_pos : float or None, optional
        Where to show Axial Slice. If None, slice is not shown.
        Should be a decimal between 0 and 1.
        Indicatesthe fractional position along the perpendicular
        axis to the slice.
        Default: 0.5

    opacity : float, optional
        Opacity of slices.
        Default: 1.0

    colorscale : string, optional
        Plotly colorscale to use to color slices.
        Default: "greys"

    invert_colorscale : bool, optional
        Whether to invert colorscale.
        Default: False

    flip_axes : ndarray
        Which axes to flip, to orient the image as RAS, which is how we
        visualize.
        For example, if the input image is LAS, use [True, False, False].
        Default: [False, False, False]

    interact : bool
        Whether to open the visualization in an interactive window.
        Default: False

    inline : bool
        Whether to embed the interactive visualization inline in a notebook.
        Only works in the notebook context. Default: False.

    Returns
    -------
    Plotly Figure object
    """
    volume = vut.load_volume(volume)
    for i, flip in enumerate(flip_axes):
        if flip:
            volume = np.flip(volume, axis=i)

    if figure is None:
        figure = make_subplots(
            rows=1, cols=1,
            specs=[[{"type": "scene"}]])

    set_layout(figure)

    for pos, axis in [(x_pos, Axes.X), (y_pos, Axes.Y), (z_pos, Axes.Z)]:
        if pos is not None:
            _draw_slice(
                figure, axis, volume, opacity=opacity,
                pos=pos, colorscale=colorscale,
                invert_colorscale=invert_colorscale)

    return _inline_interact(figure, interact, inline)


def _draw_core(sls, n_points, figure, bundle_name, indiv_profile,
               labelled_points, dimensions, flip_axes):
    fgarray = np.asarray(set_number_of_points(sls, n_points))
    fgarray = np.median(fgarray, axis=0)
    # colormap = px.colors.diverging.Portland
    # colormap = np.asarray(
    #     [[int(i) for i in c[4:-1].split(',')] for c in colormap]) / 256
    colormap = px.colors.sequential.Viridis
    colormap = np.asarray(
        [hex_to_rgb(c) for c in colormap]) / 256
    xp = np.linspace(
        np.min(indiv_profile),
        np.max(indiv_profile),
        num=len(colormap))
    line_color = np.ones((n_points, 4))
    for i in range(3):
        line_color[:, i] = np.interp(indiv_profile, xp, colormap[:, i])
    line_color_untouched = line_color.copy()
    for i in range(n_points):
        if i < n_points - 1:
            direc = fgarray[i + 1] - fgarray[i]
            direc = direc / np.linalg.norm(direc)
            light_direc = -fgarray[i] / np.linalg.norm(fgarray[i])
            direc_adjust = np.dot(direc, light_direc)
            direc_adjust = (direc_adjust + 3) / 4
        line_color[i, 0:3] = line_color[i, 0:3] * direc_adjust
    text = [None] * n_points
    for label in labelled_points:
        if label == -1:
            text[label] = str(n_points)
        else:
            text[label] = str(label)

    if flip_axes[0]:
        fgarray[:, 0] = dimensions[0] - fgarray[:, 0]
    if flip_axes[1]:
        fgarray[:, 1] = dimensions[1] - fgarray[:, 1]
    if flip_axes[2]:
        fgarray[:, 2] = dimensions[2] - fgarray[:, 2]

    figure.add_trace(
        go.Scatter3d(
            x=fgarray[:, 0],
            y=fgarray[:, 1],
            z=fgarray[:, 2],
            name=vut.display_string(bundle_name + "_core"),
            line=dict(
                width=25,
                color=line_color,
            ),
            hovertext=indiv_profile,
            hoverinfo='all',
            text=text,
            textfont=dict(size=20, family="Overpass"),
            textposition="top right",
            mode="lines+text"
        ),
        row=1, col=1
    )

    return line_color_untouched


def single_bundle_viz(indiv_profile, sft,
                      bundle, scalar_name,
                      flip_axes=[False, False, False],
                      labelled_nodes=[0, -1],
                      figure=None,
                      include_profile=False):
    """
    Visualize a single bundle in 3D with core bundle and associated profile

    Parameters
    ----------
    indiv_profile : ndarray
        A numpy array containing a tract profile for this bundle for a scalar.

    sft : Stateful Tractogram, str
        A Stateful Tractogram containing streamline information.
        If bundle is an int, the Stateful Tractogram
        must contain a bundle key in it's data_per_streamline which is a list
        of bundle `'uid'.
        Otherwise, the entire Stateful Tractogram will be used as the bundle
        for the visualization.

    bundle : str or int
        The name of the bundle to be used as the label for the plot,
        and for selection from the sft metadata.

    scalar_name : str
        The name of the scalar being used.

    flip_axes : ndarray
        Which axes to flip, to orient the image as RAS, which is how we
        visualize.
        For example, if the input image is LAS, use [True, False, False].
        Default: [False, False, False]

    labelled_nodes : list or ndarray
        Which nodes to label. -1 indicates the last node.
        Default: [0, -1]

    figure : Plotly Figure object, optional
        If provided, the visualization will be added to this Figure. Default:
        Initialize a new Figure.

    include_profile : bool, optional
        If true, also plot the tract profile. Default: False

    Returns
    -------
    Plotly Figure object
    """
    if figure is None:
        if include_profile:
            figure = make_subplots(
                rows=1, cols=2,
                specs=[[{"type": "scene"}, {"type": "scene"}]])
        else:
            figure = make_subplots(
                rows=1, cols=1,
                specs=[[{"type": "scene"}]])

    set_layout(figure)

    n_points = len(indiv_profile)
    sls, _, bundle_name, dimensions = next(vut.tract_generator(
        sft, bundle, None, n_points))

    line_color = _draw_core(
        sls, n_points, figure, bundle_name, indiv_profile,
        labelled_nodes, dimensions, flip_axes)

    if include_profile:
        _plot_profiles(
            indiv_profile, bundle_name + "_profile",
            line_color, figure, scalar_name)

    return figure



import numpy as np
import AFQ.recognition.roi as abr
import nibabel as nib
import numpy.testing as npt
import numpy as np
from scipy.spatial.distance import cdist
from AFQ.recognition.roi import (
    check_sls_with_inclusion,
    check_sl_with_inclusion,
    check_sl_with_exclusion)


streamline1 = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
streamline2 = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
roi1 = np.array([[1, 2, 3], [4, 5, 6]])
roi2 = np.array([[7, 8, 9], [10, 11, 12]])
include_rois = [roi1, roi2]
exclude_rois = [roi1]
include_roi_tols = [10, 10]
exclude_roi_tols = [1]


def test_clean_by_endpoints():
    sl = [np.array([[1, 1, 1],
                    [2, 1, 1],
                    [3, 1, 1],
                    [4, 1, 1]]),
          np.array([[1, 1, 2],
                    [2, 1, 2],
                    [3, 1, 2],
                    [4, 1, 2]]),
          np.array([[1, 1, 1],
                    [2, 1, 1],
                    [3, 1, 1]]),
          np.array([[1, 1, 1],
                    [2, 1, 1]])]

    atlas = np.zeros((20, 20, 20))

    # Targets:
    atlas[1, 1, 1] = 1
    atlas[1, 1, 2] = 2
    atlas[4, 1, 1] = 3
    atlas[4, 1, 2] = 4

    target_img_start = nib.Nifti1Image(
        np.logical_or(atlas == 1, atlas == 2).astype(np.float32), np.eye(4))
    target_img_end = nib.Nifti1Image(
        np.logical_or(atlas == 3, atlas == 4).astype(np.float32), np.eye(4))

    clean_idx_start = list(abr.clean_by_endpoints(
        sl, target_img_start, 0))
    clean_idx_end = list(abr.clean_by_endpoints(
        sl, target_img_end, -1))
    npt.assert_array_equal(np.logical_and(
        clean_idx_start, clean_idx_end), np.array([1, 1, 0, 0]))

    # If tol=1, the third streamline also gets included
    clean_idx_start = list(abr.clean_by_endpoints(
        sl, target_img_start, 0, tol=1))
    clean_idx_end = list(abr.clean_by_endpoints(
        sl, target_img_end, -1, tol=1))
    npt.assert_array_equal(np.logical_and(
        clean_idx_start, clean_idx_end), np.array([1, 1, 1, 0]))


def test_check_sls_with_inclusion():
    sls = [streamline1, streamline2]
    result = list(check_sls_with_inclusion(
        sls, include_rois, include_roi_tols))
    assert result[0][0] is True
    assert np.allclose(
        result[0][1][0], cdist(streamline1, roi1, 'sqeuclidean'))
    assert np.allclose(
        result[0][1][1], cdist(streamline1, roi2, 'sqeuclidean'))
    assert result[1][0] is False


def test_check_sl_with_inclusion_pass():
    result, dists = check_sl_with_inclusion(
        streamline1, include_rois, include_roi_tols)
    assert result is True
    assert len(dists) == 2


def test_check_sl_with_inclusion_fail():
    result, dists = check_sl_with_inclusion(
        streamline2, include_rois, include_roi_tols)
    assert result is False
    assert dists == []


def test_check_sl_with_exclusion_pass():
    result = check_sl_with_exclusion(
        streamline1, exclude_rois, exclude_roi_tols)
    assert result is False


def test_check_sl_with_exclusion_fail():
    result = check_sl_with_exclusion(
        streamline2, exclude_rois, exclude_roi_tols)
    assert result is True



import nibabel as nib
import numpy as np
from time import time
import logging

import pimms
import multiprocessing

from AFQ.tasks.decorators import as_file, as_img
from AFQ.tasks.utils import with_name
from AFQ.definitions.utils import Definition
import AFQ.tractography.tractography as aft
from AFQ.tasks.utils import get_default_args
from AFQ.definitions.image import ScalarImage
from AFQ.tractography.utils import gen_seeds

try:
    import ray
    has_ray = True
except ModuleNotFoundError:
    has_ray = False
try:
    from trx.trx_file_memmap import TrxFile
    from trx.trx_file_memmap import concatenate as trx_concatenate
    has_trx = True
except ModuleNotFoundError:
    has_trx = False

try:
    from AFQ.tractography.gputractography import gpu_track
    has_gputrack = True
except ModuleNotFoundError:
    has_gputrack = False

logger = logging.getLogger('AFQ')


def _meta_from_tracking_params(
        tracking_params, start_time,
        n_streamlines, seed, stop):
    meta_directions = {
        "det": "deterministic",
        "prob": "probabilistic"}
    meta = dict(
        TractographyClass="local",
        TractographyMethod=meta_directions.get(
            tracking_params["directions"],
            tracking_params["directions"]),
        Count=n_streamlines,
        Seeding=dict(
            ROI=seed,
            n_seeds=tracking_params["n_seeds"],
            random_seeds=tracking_params["random_seeds"]),
        Constraints=dict(ROI=stop),
        Parameters=dict(
            Units="mm",
            StepSize=tracking_params["step_size"],
            MinimumLength=tracking_params["minlen"],
            MaximumLength=tracking_params["maxlen"],
            Unidirectional=False),
        Timing=time() - start_time)
    return meta


@pimms.calc("seed")
@as_file('_desc-seed_mask.nii.gz', include_track=True)
def export_seed_mask(data_imap, tracking_params):
    """
    full path to a nifti file containing the
    tractography seed mask
    """
    seed_mask = tracking_params['seed_mask']
    seed_mask_desc = dict(source=tracking_params['seed_mask'])
    return nib.Nifti1Image(seed_mask, data_imap["dwi_affine"]), \
        seed_mask_desc


@pimms.calc("stop")
@as_file('_desc-stop_mask.nii.gz', include_track=True)
def export_stop_mask(data_imap, tracking_params):
    """
    full path to a nifti file containing the
    tractography stop mask
    """
    stop_mask = tracking_params['stop_mask']
    stop_mask_desc = dict(source=tracking_params['stop_mask'])
    return nib.Nifti1Image(stop_mask, data_imap["dwi_affine"]), \
        stop_mask_desc


@pimms.calc("stop")
def export_stop_mask_pft(pve_wm, pve_gm, pve_csf):
    """
    full path to a nifti file containing the
    tractography stop mask
    """
    return {"stop": [pve_wm, pve_gm, pve_csf]}


@pimms.calc("streamlines")
@as_file('_tractography', include_track=True)
def streamlines(data_imap, seed, stop,
                tracking_params):
    """
    full path to the complete, unsegmented tractography file

    Parameters
    ----------
    tracking_params : dict, optional
        The parameters for tracking. Default: use the default behavior of
        the aft.track function. Seed mask and seed threshold, if not
        specified, are replaced with scalar masks from scalar[0]
        thresholded to 0.2. The ``seed_mask`` and ``stop_mask`` items of
        this dict may be ``AFQ.definitions.image.ImageFile`` instances.
        If ``tracker`` is set to "pft" then ``stop_mask`` should be
        an instance of ``AFQ.definitions.image.PFTImage``.
    """
    this_tracking_params = tracking_params.copy()

    # get odf_model
    odf_model = this_tracking_params["odf_model"]
    if odf_model == "DTI":
        params_file = data_imap["dti_params"]
    elif odf_model == "CSD":
        params_file = data_imap["csd_params"]
    elif odf_model == "DKI":
        params_file = data_imap["dki_params"]
    elif odf_model == "GQ":
        params_file = data_imap["gq_params"]
    elif odf_model == "RUMBA":
        params_file = data_imap["rumba_params"]
    else:
        raise TypeError((
            f"The ODF model you gave ({odf_model}) was not recognized"))

    # get masks
    this_tracking_params['seed_mask'] = nib.load(seed).get_fdata()
    if isinstance(stop, str):
        this_tracking_params['stop_mask'] = nib.load(stop).get_fdata()
    else:
        this_tracking_params['stop_mask'] = stop

    is_trx = this_tracking_params.get("trx", False)

    num_chunks = this_tracking_params.pop("num_chunks", False)

    if num_chunks is True:
        num_chunks = multiprocessing.cpu_count() - 1

    if is_trx:
        start_time = time()
        dtype_dict = {'positions': np.float16, 'offsets': np.uint32}
        if num_chunks and num_chunks > 1:
            if not has_ray:
                raise ImportError("Ray is required to perform tractography in"
                                  "parallel, please install ray or remove the"
                                  " 'num_chunks' arg")

            @ray.remote
            class TractActor():
                def __init__(self):
                    self.TrxFile = TrxFile
                    self.aft = aft
                    self.objects = {}

                def trx_from_lazy_tractogram(self, lazyt_id, seed, dtype_dict):
                    id = self.objects[lazyt_id]
                    return self.TrxFile.from_lazy_tractogram(
                        id,
                        seed,
                        dtype_dict=dtype_dict)

                def create_lazyt(self, id, *args, **kwargs):
                    self.objects[id] = self.aft.track(*args, **kwargs)
                    return id

                def delete_lazyt(self, id):
                    if id in self.objects:
                        del self.objects[id]
            actors = [TractActor.remote() for _ in range(num_chunks)]
            object_id = 1
            tracking_params_list = []

            # random seeds case
            if isinstance(this_tracking_params.get("n_seeds"), int) and \
               this_tracking_params.get("random_seeds"):

                remainder = this_tracking_params['n_seeds'] % num_chunks
                for i in range(num_chunks):
                    # create copy of tracking params
                    copy = this_tracking_params.copy()
                    n_seeds = this_tracking_params['n_seeds']
                    copy['n_seeds'] = n_seeds // num_chunks
                    # add remainder to 1st list
                    if i == 0:
                        copy['n_seeds'] += remainder
                    tracking_params_list.append(copy)

            elif isinstance(this_tracking_params['n_seeds'], (np.ndarray,
                                                              list)):
                n_seeds = np.array(this_tracking_params['n_seeds'])
                seed_chunks = np.array_split(n_seeds, num_chunks)
                tracking_params_list = [this_tracking_params.copy() for _ in
                                        range(num_chunks)]

                for i in range(num_chunks):
                    tracking_params_list[i]['n_seeds'] = seed_chunks[i]

            else:
                seeds = gen_seeds(
                    this_tracking_params['seed_mask'],
                    this_tracking_params['seed_threshold'],
                    this_tracking_params['n_seeds'],
                    this_tracking_params['thresholds_as_percentages'],
                    this_tracking_params['random_seeds'],
                    this_tracking_params['rng_seed'],
                    data_imap["dwi_affine"])
                seed_chunks = np.array_split(seeds, num_chunks)
                tracking_params_list = [this_tracking_params.copy() for _
                                        in range(num_chunks)]
                for i in range(num_chunks):
                    tracking_params_list[i]['n_seeds'] = seed_chunks[i]

            # create lazyt inside each actor
            tasks = [ray_actor.create_lazyt.remote(object_id, params_file,
                     **tracking_params_list[i]) for i, ray_actor in
                     enumerate(actors)]
            ray.get(tasks)

            # create trx from lazyt
            tasks = [ray_actor.trx_from_lazy_tractogram.remote(object_id, seed,
                     dtype_dict=dtype_dict) for ray_actor in actors]
            sfts = ray.get(tasks)

            # cleanup objects
            tasks = [ray_actor.delete_lazyt.remote(object_id) for ray_actor in
                     actors]
            ray.get(tasks)

            sft = trx_concatenate(sfts)
        else:
            lazyt = aft.track(params_file, **this_tracking_params)
            sft = TrxFile.from_lazy_tractogram(
                lazyt,
                seed,
                dtype_dict=dtype_dict)
        n_streamlines = len(sft)

    else:
        start_time = time()
        sft = aft.track(params_file, **this_tracking_params)
        sft.to_vox()
        n_streamlines = len(sft.streamlines)

    return sft, _meta_from_tracking_params(
        tracking_params, start_time,
        n_streamlines, seed, stop)


@pimms.calc("streamlines")
def custom_tractography(import_tract=None):
    """
    full path to the complete, unsegmented tractography file

    Parameters
    ----------
    import_tract : dict or str or None, optional
        BIDS filters for inputing a user made tractography file,
        or a path to the tractography file. If None, DIPY is used
        to generate the tractography.
        Default: None
    """
    if not isinstance(import_tract, str):
        raise TypeError(
            "import_tract must be"
            + " either a dict or a str")
    return import_tract


@pimms.calc("streamlines")
@as_file('_tractography', include_track=True)
def gpu_tractography(data_imap, tracking_params, seed, stop,
                     tractography_ngpus=0, chunk_size=100000):
    """
    full path to the complete, unsegmented tractography file

    Parameters
    ----------
    tractography_ngpus : int, optional
        Number of GPUs to use in tractography. If non-0,
        this algorithm is used for tractography,
        https://github.com/dipy/GPUStreamlines
        Default: 0
    chunk_size : int, optional
        Chunk size for GPU tracking.
        Default: 100000
    """
    start_time = time()
    sft = gpu_track(
        data_imap["data"], data_imap["gtab"],
        nib.load(seed), nib.load(stop),
        tracking_params["odf_model"],
        tracking_params["seed_threshold"],
        tracking_params["stop_threshold"],
        tracking_params["thresholds_as_percentages"],
        tracking_params["max_angle"], tracking_params["step_size"],
        tracking_params["n_seeds"],
        tracking_params["random_seeds"],
        tracking_params["rng_seed"],
        tractography_ngpus,
        chunk_size)

    return sft, _meta_from_tracking_params(
        tracking_params, start_time,
        sft, seed, stop)


def get_tractography_plan(kwargs):
    if "tracking_params" in kwargs\
            and not isinstance(kwargs["tracking_params"], dict):
        raise TypeError(
            "tracking_params a dict")

    tractography_tasks = with_name([
        export_seed_mask, export_stop_mask, streamlines])

    # use GPU accelerated tractography if asked for
    if "tractography_ngpus" in kwargs and kwargs["tractography_ngpus"] != 0:
        if not has_gputrack:
            raise ImportError("Please install from ghcr.io/nrdg/pyafq_gpu"
                              " docker file to use gpu-accelerated"
                              "tractography")
        tractography_tasks["streamlines_res"] = gpu_tractography
    # use imported tractography if given
    if "import_tract" in kwargs and kwargs["import_tract"] is not None:
        tractography_tasks["streamlines_res"] = custom_tractography

    # determine reasonable defaults
    best_scalar = kwargs["scalars"][0]
    fa_found = False
    for scalar in kwargs["scalars"]:
        if isinstance(scalar, str):
            if "fa" in scalar:
                best_scalar = scalar
                fa_found = True
                break
        else:
            if "fa" in scalar.get_name():
                best_scalar = scalar
                fa_found = True
                break
    if not fa_found:
        logger.warning(
            "FA not found in list of scalars, will use first scalar"
            " for the seed and stop mask and visualizations"
            " unless these are also specified")
    kwargs["best_scalar"] = best_scalar

    default_tracking_params = get_default_args(aft.track)

    # Replace the defaults only for kwargs for which a non-default value
    # was given:
    if "tracking_params" in kwargs:
        for k in kwargs["tracking_params"]:
            default_tracking_params[k] = kwargs["tracking_params"][k]

    kwargs["tracking_params"] = default_tracking_params
    kwargs["tracking_params"]["odf_model"] =\
        kwargs["tracking_params"]["odf_model"].upper()
    if kwargs["tracking_params"]["seed_mask"] is None:
        kwargs["tracking_params"]["seed_mask"] = ScalarImage(
            kwargs["best_scalar"])
        kwargs["tracking_params"]["seed_threshold"] = 0.2
        logger.info((
            "No seed mask given, using FA (or first scalar if none are FA)"
            "thresholded to 0.2"))
    if kwargs["tracking_params"]["stop_mask"] is None:
        kwargs["tracking_params"]["stop_mask"] = ScalarImage(
            kwargs["best_scalar"])
        kwargs["tracking_params"]["stop_threshold"] = 0.2
        logger.info((
            "No stop mask given, using FA (or first scalar if none are FA)"
            "thresholded to 0.2"))

    stop_mask = kwargs["tracking_params"]['stop_mask']
    seed_mask = kwargs["tracking_params"]['seed_mask']

    if kwargs["tracking_params"]["tracker"] == "pft":
        probseg_funcs = stop_mask.get_image_getter("tractography")
        tractography_tasks["wm_res"] = pimms.calc("pve_wm")(probseg_funcs[0])
        tractography_tasks["gm_res"] = pimms.calc("pve_gm")(probseg_funcs[1])
        tractography_tasks["csf_res"] = pimms.calc("pve_csf")(probseg_funcs[2])
        tractography_tasks["export_stop_mask_res"] = \
            export_stop_mask_pft
    elif isinstance(stop_mask, Definition):
        tractography_tasks["export_stop_mask_res"] = pimms.calc("stop")(
            as_file('_desc-stop_mask.nii.gz', include_track=True)(
                stop_mask.get_image_getter("tractography")))

    if isinstance(seed_mask, Definition):
        tractography_tasks["export_seed_mask_res"] = pimms.calc("seed")(
            as_file('_desc-seed_mask.nii.gz', include_track=True)(
                seed_mask.get_image_getter("tractography")))

    return pimms.plan(**tractography_tasks)


def _gen_seeds(n_seeds, params_file, seed_mask=None, seed_threshold=0,
               thresholds_as_percentages=False,
               random_seeds=False, rng_seed=None):
    if isinstance(params_file, str):
        params_img = nib.load(params_file)
    else:
        params_img = params_file

    affine = params_img.affine

    return gen_seeds(seed_mask, seed_threshold, n_seeds,
                     thresholds_as_percentages,
                     random_seeds, rng_seed, affine)


import os.path as op

import pytest

import numpy as np
import numpy.testing as npt

import nibabel as nib
import dipy.data as dpd
import dipy.data.fetcher as fetcher
from dipy.stats.analysis import afq_profile
from dipy.io.stateful_tractogram import StatefulTractogram, Space

import AFQ.data.fetch as afd
import AFQ.registration as reg
import AFQ.recognition.cleaning as abc
from AFQ.recognition.recognize import recognize


dpd.fetch_stanford_hardi()
hardi_dir = op.join(fetcher.dipy_home, "stanford_hardi")
hardi_fdata = op.join(hardi_dir, "HARDI150.nii.gz")
hardi_img = nib.load(hardi_fdata)
hardi_fbval = op.join(hardi_dir, "HARDI150.bval")
hardi_fbvec = op.join(hardi_dir, "HARDI150.bvec")
file_dict = afd.read_stanford_hardi_tractography()
reg_template = afd.read_mni_template()
mapping = reg.read_mapping(
    file_dict['mapping.nii.gz'],
    hardi_img,
    reg_template)
streamlines = file_dict['tractography_subsampled.trk']
tg = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
tg.to_vox()
streamlines = tg.streamlines
templates = afd.read_templates()
cst_r_curve_ref = StatefulTractogram([[
    [4.09933186, -27.86049652, -8.57825184],
    [4.18979505, -27.30110527, -7.7542993 ],
    [4.31947752, -26.78867352, -6.90792487],
    [4.48837897, -26.32320413, -6.03912657],
    [4.74019388, -25.95165819, -5.14551435],
    [4.99200908, -25.58011067, -4.25190286],
    [5.29724114, -25.17955789, -3.39071481],
    [5.70181712, -24.72099485, -2.60239253],
    [6.10747528, -24.18646430, -1.8638705 ],
    [6.56050014, -23.51795578, -1.27745605]]],
    reg_template, Space.RASMM)

bundles = {'Left Corticospinal': {
                    'include': [
                        templates['CST_roi1_L'],
                        templates['CST_roi2_L']],
                    'prob_map': templates['CST_L_prob_map'],
                    'cross_midline': None},
           'Right Corticospinal': {
                    'include': [
                        templates['CST_roi1_R'],
                        templates['CST_roi2_R']],
                    'prob_map': templates['CST_R_prob_map'],
                    "curvature": {
                        "sft": cst_r_curve_ref,
                        "cut": True, "thresh": 45},
                    'cross_midline': None}}

def test_segment():
    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles,
        reg_template)

    # We asked for 2 fiber groups:
    npt.assert_equal(len(fiber_groups), 2)
    # Here's one of them:
    CST_R_sl = fiber_groups['Right Corticospinal']
    # Let's make sure there are streamlines in there:
    npt.assert_(len(CST_R_sl) > 0)
    # Calculate the tract profile for a volume of all-ones:
    tract_profile = afq_profile(
        np.ones(nib.load(hardi_fdata).shape[:3]),
        CST_R_sl.streamlines, np.eye(4))
    npt.assert_almost_equal(tract_profile, np.ones(100))

    clean_sl = abc.clean_bundle(CST_R_sl)
    npt.assert_equal(len(clean_sl), len(CST_R_sl))


@pytest.mark.nightly
def test_segment_no_prob():
    # What if you don't have probability maps?
    bundles_no_prob = {
        'Left Corticospinal': {
            'include': [
                templates['CST_roi1_L'],
                templates['CST_roi2_L']],
            'cross_midline': False},
        'Right Corticospinal': {
            'include': [
                templates['CST_roi1_R'],
                templates['CST_roi2_R']],
            'cross_midline': False}}

    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles_no_prob,
        reg_template)

    # This condition should still hold
    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['Right Corticospinal']) > 0)


def test_segment_return_idx():
    # Test with the return_idx kwarg set to True:
    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles,
        reg_template,
        return_idx=True)

    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['Right Corticospinal']['sl']) > 0)
    npt.assert_(len(fiber_groups['Right Corticospinal']['idx']) > 0)


@pytest.mark.nightly
def test_segment_clip_edges_api():
    # Test with the clip_edges kwarg set to True:
    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles,
        reg_template,
        clip_edges=True)
    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['Right Corticospinal']) > 0)


def test_segment_reco():
    # get bundles for reco method
    bundles_reco = afd.read_hcp_atlas(16)
    bundle_names = ['CST_R', 'CST_L']
    for key in list(bundles_reco):
        if key not in bundle_names:
            bundles_reco.pop(key, None)

    # Try recobundles method
    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles_reco,
        reg_template,
        rng=np.random.RandomState(seed=8))

    # This condition should still hold
    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['CST_R']) > 0)


def test_exclusion_ROI():
    slf_bundle = {
        'Left Superior Longitudinal': {
            'include': [
                templates['SLF_roi1_L'],
                templates['SLF_roi2_L']],
            'cross_midline': None}}

    # tractogram where 1 streamline goes through include ROIs only
    # and the other goes through both include and exclude ROIs
    slf_tg = StatefulTractogram(
        np.asarray(
            [
                [
                    [8, 53, 39], [8, 50, 39], [8, 45, 39],
                    [30, 41, 61], [28, 61, 38]],
                [
                    [8, 53, 39], [8, 50, 39], [8, 45, 39],
                    [30, 41, 62], [20, 44, 34]]
            ]).astype(float),
        hardi_img, Space.VOX)
    fiber_groups, _ = recognize(
        slf_tg,
        nib.load(hardi_fdata),
        mapping,
        slf_bundle,
        reg_template)
     
    npt.assert_equal(len(fiber_groups["Left Superior Longitudinal"]), 2)

    slf_bundle['Left Superior Longitudinal']['exclude'] =\
        [templates["SLFt_roi2_L"]]

    fiber_groups, _ = recognize(
        slf_tg,
        nib.load(hardi_fdata),
        mapping,
        slf_bundle,
        reg_template)

    npt.assert_equal(len(fiber_groups["Left Superior Longitudinal"]), 1)


def test_segment_sampled_streamlines():
    fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles,
        reg_template)

    # Already using a subsampled tck
    # the Right Corticospinal has two streamlines and
    # Left Corticospinal has none
    npt.assert_(0 < len(fiber_groups['Right Corticospinal']))

    # number of streamlines to sample
    nb_streamlines = int(len(tg)*0.8)

    # sample and segment streamlines
    sampled_fiber_groups, _ = recognize(
        tg,
        nib.load(hardi_fdata),
        mapping,
        bundles,
        reg_template,
        nb_streamlines=nb_streamlines)

    # sampled streamlines should be subset of the original streamlines
    npt.assert_(
        np.all(
            np.isin(
                sampled_fiber_groups['Right Corticospinal'].streamlines._data,
                fiber_groups['Right Corticospinal'].streamlines._data
            )
        )
    )

    # expect the number of resulting streamlines to be more than 0 but less
    # than default; given that random sample and given there are only two
    # streamlines less than equal
    npt.assert_(0 <= len(
        sampled_fiber_groups['Right Corticospinal']) <= len(
            fiber_groups['Right Corticospinal']))



"""
========================================
Using Subject Space ROIs from Freesurfer
========================================

An example using the AFQ API to find bundles
as defined by endpoint ROIs from freesurfer.
This example can be modified to work with ROIs
in subject space from pipelines other than freesurfer.
"""
import os.path as op

import nibabel as nib
import plotly
import numpy as np

from AFQ.api.group import GroupAFQ
import AFQ.data.fetch as afd
from AFQ.definitions.image import RoiImage
import AFQ.api.bundle_dict as abd

##########################################################################
# Get some example data
# ---------------------
#
# Retrieves High angular resolution diffusion imaging (HARDI) dataset from
# Stanford's Vista Lab
#
#   see https://purl.stanford.edu/ng782rw8378 for details on dataset.
#
# The data for the first subject and first session are downloaded locally
# (by default into the users home directory) under:
#
#   ``.dipy/stanford_hardi/``
#
# Anatomical data (``anat``) and Diffusion-weighted imaging data (``dwi``) are
# then extracted, formatted to be BIDS compliant, and placed in the AFQ
# data directory (by default in the users home directory) under:
#
#   ``AFQ_data/stanford_hardi/``
#
# This data represents the required preprocessed diffusion data necessary for
# intializing the GroupAFQ object (which we will do next)
#
# The clear_previous_afq is used to remove any previous runs of the afq object
# stored in the AFQ_data/stanford_hardi/ BIDS directory. Set it to None if
# you want to use the results of previous runs. Setting it to "track"
# as here will only clear derivatives that depend on the tractography stage
# (i.e., bundle delination and tract profile calculation),
# as well as the tractography itself, to save time on recomputation.
# If you want to only clear derivatives that depend on bundle delineation,
# and keep the tractography, you can set clear_previous_afq to
# "recog" instead.

afd.organize_stanford_data(clear_previous_afq="track")

##########################################################################
# Generate left thalamus ROI from freesurfer segmentation file
# ------------------------------------------------------------
# 1. Load the segmentation file that was generated by Freesurfer for
#    the specific subject.
# 2. Identify the left thalamus within the file, which has the label
#    number 41
# 3. Create a Nifti image representing the left thalamus ROI:
#    - Assign a value of 1 to the voxels that Freesurfer
#      has labeled as 41 (i.e., the left thalamus).
#    - Assign a value of 0 to all other voxels.
# This binary mask format is the expected input for pyAFQ when
# dealing with subject space ROIs. If it's already in binary format,
# there is no need to do this step.

freesurfer_subject_folder = op.join(
    afd.afq_home, "stanford_hardi",
    "derivatives", "freesurfer",
    "sub-01", "ses-01",
    "anat")

seg_file = nib.load(op.join(
    freesurfer_subject_folder, "sub-01_ses-01_seg.nii.gz"))
left_thal = seg_file.get_fdata() == 41
nib.save(
    nib.Nifti1Image(
        left_thal.astype(np.float32),
        seg_file.affine),
    op.join(
        freesurfer_subject_folder,
        "sub-01_ses-01_desc-leftThal_mask.nii.gz"))

# Fetch LV1 ROI
# which was already generated using the process above
afd.fetch_stanford_hardi_lv1()

##########################################################################
# Set tractography parameters (optional)
# ---------------------
# We make this tracking_params which we will pass to the GroupAFQ object
# which specifies that we want 10,000 seeds randomly distributed
# only within the endpoint ROIs and not throughout the white matter.
# This is controlled by passing
# `"seed_mask": RoiImage()` in the `tracking_params` dict.
#
# We only do this to make this example faster and consume less space.

tracking_params = dict(n_seeds=10000,
                       random_seeds=True,
                       rng_seed=42,
                       seed_mask=RoiImage(use_endpoints=True))

#############################################################################
# Define custom `BundleDict` object
# --------------------------------
# In a typical `BundleDict` object, ROIs are passed as paths to Nifti files.
# Here, we define ROIs as dictionaries instead, containing BIDS filters.
# Then pyAFQ can find the respective ROI for each subject and session.

bundles = abd.BundleDict({
    "L_OR": {
        "start": {
            "scope": "freesurfer",
            "suffix": "mask",
            "desc": "leftThal"},
        "end": {
            "scope": "freesurfer",
            "suffix": "anat",
            "desc": "LV1"
        },
        "cross_midline": False,
        "space": "subject"
    }})

##########################################################################
# Initialize a GroupAFQ object:
# -------------------------
#
# Creates a GroupAFQ object, that encapsulates tractometry,
# passing in our custom bundle info. Then we run the pipeline
# and generate a visualization of the bundle we found.

myafq = GroupAFQ(
    bids_path=op.join(afd.afq_home, 'stanford_hardi'),
    preproc_pipeline='vistasoft',
    tracking_params=tracking_params,
    bundle_info=bundles)

bundle_html = myafq.export("indiv_bundles_figures")
plotly.io.show(bundle_html["01"]["L_OR"])



import os
import os.path as op

import numpy as np
import nibabel as nib

from dipy.reconst import dti
from scipy.special import gamma

import AFQ.utils.models as ut

__all__ = ["fit_dti", "predict"]


def noise_from_b0(data, gtab, bvals, mask=None, b0_threshold=50):
    """
    Coped from vistasoft's dtiComputeImageNoise
    https://github.com/vistalab/vistasoft
    And converted to python...

    Parameters
    ----------
    data_files : str or list
        Files containing DWI data. If this is a str, that's the full path to a
        single file. If it's a list, each entry is a full path.
    bval_files : str or list
        Equivalent to `data_files`.
    mask : ndarray, optional
        Binary mask, set to True or 1 in voxels to be processed.
        Default: Process all voxels.
    b0_threshold : float
    """
    # make all-inclusive mask if None is provided
    if mask is None:
        mask = np.ones(data.shape[:3])

    # Number of volumes in the dw dataset
    num_vols = data.shape[3]

    # Get brainmask indices
    brain_inds = (mask > 0)

    # preallocate a 2d array
    # The first dimension is the number of volumes
    # and the 2nd is each voxel (within the brain mask)
    masked_data = np.zeros((num_vols, np.sum(brain_inds)))
    # Loop over the volumes and assign the voxels within the brain mask
    # to masked_data
    for i in range(num_vols):
        tmp = data[:, :, :, i]
        masked_data[i, :] = tmp[brain_inds]

    # Find which volumes are b=0
    b0_inds = (bvals > b0_threshold)
    n = len(b0_inds)
    # Pull out the b=0 volumes
    b0_data = masked_data[b0_inds, :]
    # Calculate the median of the standard deviation. We do not think that
    # this needs to be rescaled. Henkelman et al. (1985) suggest that this
    # aproaches the true noise as the signal increases.
    sigma = np.median(np.std(b0_data, axis=1, ddof=1))

    # std of a sample underestimates sigma
    # (see http://nbviewer.ipython.org/4287207/)
    # This can be very big for small n (e.g., 20# for n=2)
    # We can compute the underestimation bias:
    bias = sigma * (1. - np.sqrt(2. / (n - 1))
                    * (gamma(n / 2.) / gamma((n - 1) / 2.)))

    # and correct for it:
    return sigma + bias


def _fit(gtab, data, mask=None, sigma=None):
    if sigma is None:
        dtimodel = dti.TensorModel(gtab)
    else:
        dtimodel = dti.TensorModel(
            gtab,
            fit_method="RT",
            sigma=sigma)
    return dtimodel.fit(data, mask=mask)


def fit_dti(data_files, bval_files, bvec_files, mask=None,
            out_dir=None, file_prefix=None, b0_threshold=50):
    """
    Fit the DTI model using default settings, save files with derived maps.

    Parameters
    ----------
    data_files : str or list
        Files containing DWI data. If this is a str, that's the full path to a
        single file. If it's a list, each entry is a full path.
    bval_files : str or list
        Equivalent to `data_files`.
    bvec_files : str or list
        Equivalent to `data_files`.
    mask : ndarray, optional
        Binary mask, set to True or 1 in voxels to be processed.
        Default: Process all voxels.
    out_dir : str, optional
        A full path to a directory to store the maps that get computed.
        Default: maps get stored in the same directory as the last DWI file
        in `data_files`.
    b0_threshold : float

    Returns
    -------
    file_paths : dict
        A dict with the derived maps that were computed and full-paths
        to the files containing these maps.

    Notes
    -----
    Maps that are calculated: FA, MD, AD, RD
    """
    img, data, gtab, mask = ut.prepare_data(data_files, bval_files,
                                            bvec_files, mask=mask,
                                            b0_threshold=b0_threshold)

    # In this case, we dump the fit object
    dtf = _fit(gtab, data, mask=mask)
    FA, MD, AD, RD, params = dtf.fa, dtf.md, dtf.ad, dtf.rd, dtf.model_params

    maps = [FA, MD, AD, RD, params]
    names = ['FA', 'MD', 'AD', 'RD', 'params']

    if out_dir is None:
        if isinstance(data_files, list):
            out_dir = op.join(op.split(data_files[0])[0], 'dti')
        else:
            out_dir = op.join(op.split(data_files)[0], 'dti')
    if file_prefix is None:
        file_prefix = ''

    if not op.exists(out_dir):
        os.makedirs(out_dir)

    aff = img.affine
    file_paths = {}
    for m, n in zip(maps, names):
        file_paths[n] = op.join(out_dir, file_prefix + 'dti_%s.nii.gz' % n)
        nib.save(nib.Nifti1Image(m, aff), file_paths[n])

    return file_paths


def predict(params_file, gtab, S0_file=None, out_dir=None):
    """
    Create a signal prediction from DTI params.

    Parameters
    ----------
    params_file : str
        Full path to a file with parameters saved from a DKI fit

    gtab : GradientTable object
        The gradient table to predict for

    S0_file : str
        Full path to a nifti file that contains S0 measurements to incorporate
        into the prediction. If the file contains 4D data, the volumes that
        contain the S0 data must be the same as the gtab.b0s_mask.

    Returns
    -------
    fname : str
        The name of the nifti file with saved predictions.

    """
    if out_dir is None:
        out_dir = op.join(op.split(params_file)[0])

    if S0_file is None:
        S0 = 100
    else:
        S0 = nib.load(S0_file).get_fdata()
        # If the S0 data is 4D, we assume it comes from an acquisition that had
        # B0 measurements in the same volumes described in the gtab:
        if len(S0.shape) == 4:
            S0 = np.mean(S0[..., gtab.b0s_mask], -1)
        # Otherwise, we assume that it's already a 3D volume, and do nothing

    img = nib.load(params_file)
    params = img.get_fdata()
    pred = dti.tensor_prediction(params, gtab, S0=S0)
    fname = op.join(out_dir, 'dti_prediction.nii.gz')
    nib.save(nib.Nifti1Image(pred, img.affine), fname)

    return fname



"""
=====================================================
How to add new bundles into pyAFQ (SLF 1/2/3 Example)
=====================================================

pyAFQ is designed to be customizable and extensible. This example shows how you
can customize it to define a new bundle based on a definition of waypoint and
endpoint ROIs of your design.

In this case, we add sub-bundles of the superior longitudinal fasciculus,
based on work by Sami et al [1]_.

We start by importing some of the components that we need for this example and
fixing the random seed for reproducibility
"""

import os.path as op
import plotly
import numpy as np
import shutil

from AFQ.api.group import GroupAFQ
import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd
from AFQ.definitions.image import ImageFile, RoiImage
import wget
import os
np.random.seed(1234)


#############################################################################
# Get dMRI data
# ---------------
# We will analyze eight subject from the Healthy Brain Network Processed Open
# Diffusion Derivatives dataset (HBN-POD2) [2]_, [3]_. We'll use a fetcher to
# get preprocessed dMRI data for eight of the >2,000 subjects in that study. The
# data gets organized into a BIDS-compatible format in the `~/AFQ_data/HBN`
# folder. These 12 subjects have very high quality data.
# The fether returns this directory as study_dir:

_, study_dir = afd.fetch_hbn_preproc([
    'NDARKP893TWU',
    'NDAREP505XAD',
    'NDARKT540ZW0',
    'NDARAG340ERT',
    'NDAREM757NBG',
    'NDARLL894HC3',
    'NDARFY525TL2',
    'NDARKV461KGZ',
    'NDARUC851WHU',
    'NDARMJ333WJM',
    'NDARJG687YYX',
    'NDARJA157YB3',
])

#############################################################################
# Get ROIs and save to disk
# --------------------------------
# The goal of this tutorial is to demostrate how to segment new pathways based
# on ROIs that are saved to disk. In principle, ROIs can be a) files created by
# the user and saved to the local disk, b) files stored somewhere on the internet
# (as is the case here) or c) Files that are accessed with a fetcher. In this
# example we download these files from the MATLAB AFQ website, but this code could
# be commented out and paths could be used to local ROIs on disk

roi_urls = ['https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/MFgL.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/MFgR.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/PaL.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/PaR.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/PrgL.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/PrgR.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/SFgL.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/SFgR.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/SLFt_roi2_L.nii.gz',
            'https://github.com/yeatmanlab/AFQ/raw/c762ca4c393f2105d4f444c44d9e4b4702f0a646/SLF123/ROIs/SLFt_roi2_R.nii.gz']

# We proceed to download the files. First, we define and create the directory
# for the template ROIs. In the code below, ``op.expanduser("~")`` expands the
# user's home directory into the full path and ``op.join`` joins these paths,
# to make the path `~/AFQ_data/SLF_ROIs/`

template_dir = op.join(
    op.expanduser("~"),
    'AFQ_data/SLF_ROIs/')
os.makedirs(template_dir, exist_ok=True)

# The `wget` Python library works like the `wget` unix command and downloads
# each file into the directory created just above.

for roi_url in roi_urls:
    wget.download(roi_url, template_dir)


#############################################################################
# Define custom `BundleDict` object
# ---------------------------------
# A `BundleDict` is a custom object that holds information about "include" and
# "exclude" ROIs, as well as endpoint ROIs, and whether the bundle crosses the
# midline. In this case, the ROIs are all defined in the MNI template space that
# is used as the default template space in pyAFQ, but, in principle, other
# template spaces could be used. In this example, we provide paths to the ROIs
# to populate the `BundleDict`, but we could also provide already-loaded nifti
# objects, as demonstrated in other examples.

bundles = abd.BundleDict({
    "L_SLF1": {
        "include": [
            template_dir + 'SFgL.nii.gz',
            template_dir + 'PaL.nii.gz'],
        "exclude": [
            template_dir + 'SLFt_roi2_L.nii.gz'],

        "cross_midline": False,

        "mahal": {
            "clean_rounds": 20,
            "length_threshold": 4,
            "distance_threshold": 2}
    },
    "L_SLF2": {
        "include": [
            template_dir + 'MFgL.nii.gz',
            template_dir + 'PaL.nii.gz'],
        "exclude": [
            template_dir + 'SLFt_roi2_L.nii.gz'],

        "cross_midline": False,

        "mahal": {
            "clean_rounds": 20,
            "length_threshold": 4,
            "distance_threshold": 2}
    },
    "L_SLF3": {
        "include": [
            template_dir + 'PrgL.nii.gz',
            template_dir + 'PaL.nii.gz'],
        "exclude": [
            template_dir + 'SLFt_roi2_L.nii.gz'],

        "cross_midline": False,

        "mahal": {
            "clean_rounds": 20,
            "length_threshold": 4,
            "distance_threshold": 2}
    }
})

#############################################################################
# Custom bundle definitions such as the SLF or OR, and the standard BundleDict
# can be combined through addition. To get both the SLF and the standard
# bundles, we would execute the following code::
#
#     bundles = bundles + abd.default18_bd()
#
# In this case, we will skip this and generate just the SLF.

#############################################################################
# Define GroupAFQ object
# ----------------------
# HBN POD2 have been processed with qsiprep [4]_. This means that a brain mask
# has already been computed for them.
#
# For tractography, we use CSD-based probabilistic tractography seeding
# extensively (`n_seeds=4` means 81 seeds per voxel!), but only within the ROIs
# and not throughout the white matter. This is controlled by passing
# `"seed_mask": RoiImage()` in the `tracking_params` dict. The custom bundles
# are passed as `bundle_info=bundles`. The call to `my_afq.export_all()`
# initiates the pipeline.

brain_mask_definition = ImageFile(
    suffix="mask",
    filters={'desc': 'brain',
             'space': 'T1w',
             'scope': 'qsiprep'})

my_afq = GroupAFQ(
    bids_path=study_dir,
    preproc_pipeline="qsiprep",
    output_dir=op.join(study_dir, "derivatives", "afq_slf"),
    brain_mask_definition=brain_mask_definition,
    tracking_params={"n_seeds": 4,
                     "directions": "prob",
                     "odf_model": "CSD",
                     "seed_mask": RoiImage()},
    segmentation_params={"parallel_segmentation": {"engine": "serial"}},
    bundle_info=bundles)

# If you want to redo different stages you can use the `clobber` method.
# The options for dependent_on are 'track' (to start over from tractography)
# or 'recog' to start over from bundle recognition. For example, to redo everying
# related  to bundle recognition: `my_afq.clobber(dependent_on='recog')`.
# This is useful when changing something about how the bundles are recognized.
# For example, the cleaning parameters.

my_afq.clobber(dependent_on='recog')

my_afq.export_all()

#############################################################################
# Visualize a montage
# ----------------------
# One way to examine the output of the pyAFQ pipeline is by creating a montage
# of images of a particular bundle across a group of participants. In the montage function
# the first input refers to a key in the bundlediect and the second gives the layout
# of the figure (eg. 3 rows 4 columns) and finally is the view.

montage = my_afq.group_montage(
    "L_SLF1", (3, 4), "Sagittal", "left", slice_pos=0.5)
montage = my_afq.group_montage(
    "L_SLF2", (3, 4), "Sagittal", "left", slice_pos=0.5)
montage = my_afq.group_montage(
    "L_SLF3", (3, 4), "Sagittal", "left", slice_pos=0.5)

#############################################################################
# Interactive bundle visualization
# --------------------------------
# Another way to examine the outputs is to export the individual bundle
# figures, which show the streamlines, as well as the ROIs used to define the
# bundle. This is an html file, which contains an interactive figure that can
# be navigated, zoomed, rotated, etc.

bundle_html = my_afq.export("all_bundles_figure")

#############################################################################
# References
# ----------
# .. [1] Romi Sagi, J.S.H. Taylor, Kyriaki Neophytou, Tamar Cohen,
#     Brenda Rapp, Kathleen Rastle, Michal Ben-Shachar.
#     White matter associations with spelling performance
#     https://doi.org/10.21203/rs.3.rs-3282349/v1
#
# .. [2] Alexander LM, Escalera J, Ai L, et al. An open resource for
#     transdiagnostic research in pediatric mental health and learning
#     disorders. Sci Data. 2017;4:170181.
#
# .. [3] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and quality
#     controlled resource for pediatric brain white-matter research. Scientific
#     Data. 2022;9(1):1-27.
#
# .. [4] Cieslak M, Cook PA, He X, et al. QSIPrep: an integrative platform for
#     preprocessing and reconstructing diffusion MRI data. Nat Methods.
#     2021;18(7):775-778.
